# Copyright (c) 2020 6WIND S.A.
# SPDX-License-Identifier: BSD-3-Clause

import grp
import logging
import os
import pwd
import unittest

import libyang

import sysrepo


YANG_FILE = os.path.join(
    os.path.dirname(os.path.dirname(__file__)), "examples/sysrepo-example.yang"
)
sysrepo.configure_logging(stderr_level=logging.ERROR)


# ------------------------------------------------------------------------------
class ConnectionTest(unittest.TestCase):
    def test_conn_connect(self):
        conn = sysrepo.SysrepoConnection()
        conn.disconnect()

    def test_conn_connect_ctxmgr(self):
        with sysrepo.SysrepoConnection():
            pass

    def test_conn_start_session(self):
        with sysrepo.SysrepoConnection() as conn:
            sess = conn.start_session()
            self.assertEqual(sess.get_datastore(), "running")
            sess.stop()

    def test_conn_start_session_ctxmgr(self):
        with sysrepo.SysrepoConnection() as conn:
            with conn.start_session() as sess:
                self.assertEqual(sess.get_datastore(), "running")

    def test_conn_start_session_operational(self):
        with sysrepo.SysrepoConnection() as conn:
            with conn.start_session("operational") as sess:
                self.assertEqual(sess.get_datastore(), "operational")

    def test_conn_install_remove_modules(self):
        YANG_FILE2 = os.path.join(
            os.path.dirname(os.path.dirname(__file__)), "examples/sysrepo-example2.yang"
        )
        with sysrepo.SysrepoConnection() as conn:
            filepaths = {YANG_FILE: ["turbo"], YANG_FILE2: ["turbo2"]}
            # filepaths = {YANG_FILE2: ["turbo2"]}
            conn.install_modules(filepaths)

        with sysrepo.SysrepoConnection() as conn:
            modules = ["sysrepo-example", "sysrepo-example2"]
            conn.remove_modules(modules)

    def test_conn_install_remove_module(self):
        with sysrepo.SysrepoConnection() as conn:
            conn.install_module(YANG_FILE, enabled_features=["turbo"])
        with sysrepo.SysrepoConnection() as conn:
            conn.remove_module("sysrepo-example")
        with sysrepo.SysrepoConnection() as conn:
            with conn.get_ly_ctx() as ctx:
                with self.assertRaises(libyang.LibyangError):
                    ctx.get_module("sysrepo-example")

    def test_conn_enable_module_feature(self):
        with sysrepo.SysrepoConnection() as conn:
            conn.install_module(YANG_FILE)
            conn.enable_module_feature("sysrepo-example", "turbo")
            with conn.start_session("operational") as sess:
                data = sess.get_data("/ietf-yang-library:*")
                data = data["yang-library"]["module-set"]
                data = next(iter(data))
                data = data["module"]
                data = [x for x in data if x["name"] == "sysrepo-example"][0]
                self.assertIn("feature", data)
            conn.remove_module("sysrepo-example")

    def test_conn_get_module_infos(self):
        with sysrepo.SysrepoConnection() as conn:
            conn.install_module(YANG_FILE)
            owner, group, perm = conn.get_module_ds_access("sysrepo-example")
            self.assertEqual(pwd.getpwnam(owner).pw_uid, os.geteuid())
            self.assertEqual(grp.getgrnam(group).gr_gid, os.getegid())
            self.assertEqual(perm, 0o600)

    def test_conn_start_session_with_printed_context(self):
        with sysrepo.SysrepoConnection(not_printed=False) as conn:
            sess = conn.start_session()
            self.assertEqual(sess.get_datastore(), "running")
            sess.stop()



# Copyright (c) 2020 6WIND S.A.
# SPDX-License-Identifier: BSD-3-Clause

import logging
import os
import threading
import time
import types
import unittest
from unittest.mock import ANY

import sysrepo


YANG_FILE = os.path.join(
    os.path.dirname(os.path.dirname(__file__)), "examples/sysrepo-example.yang"
)
sysrepo.configure_logging(stderr_level=logging.ERROR)


# ------------------------------------------------------------------------------
class SessionTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.conn = sysrepo.SysrepoConnection()
        cls.conn.install_module(YANG_FILE, enabled_features=["turbo"])

    @classmethod
    def tearDownClass(cls):
        cls.conn.remove_module("sysrepo-example")
        cls.conn.disconnect()

    def test_session_switch_ds(self):
        with self.conn.start_session("running") as sess:
            self.assertEqual(sess.get_datastore(), "running")
            sess.switch_datastore("startup")
            self.assertEqual(sess.get_datastore(), "startup")
            sess.switch_datastore("operational")
            self.assertEqual(sess.get_datastore(), "operational")

    def test_session_bad_datastore(self):
        with self.assertRaises(ValueError):
            self.conn.start_session("invalid")

    def test_session_get_ly_ctx(self):
        with self.conn.start_session() as sess:
            with sess.get_ly_ctx() as ctx:
                mod = ctx.get_module("sysrepo-example")
            self.assertTrue(mod.implemented)

    def test_session_get_data(self):
        with self.conn.start_session("operational") as sess:
            data = sess.get_data("/ietf-yang-library:*")
            self.assertIn("yang-library", data)
            data = data["yang-library"]
            self.assertIn("module-set", data)
            data = data["module-set"]
            self.assertIsInstance(data, list)
            data = next(iter(data))
            self.assertIn("module", data)
            data = data["module"]
            self.assertIsInstance(data, list)
            modinfo = {
                "name": "sysrepo-example",
                "namespace": "n",
                "feature": ["turbo"],
                "location": ANY,
            }
            self.assertIn(modinfo, data)

    MOD_XPATH = "/ietf-yang-library:yang-library/module-set/module[name=%r]/name"
    MODS_XPATH = "/ietf-yang-library:yang-library/module-set/module/name"

    def test_session_get_item_not_found(self):
        with self.conn.start_session("operational") as sess:
            with self.assertRaises(sysrepo.SysrepoNotFoundError):
                sess.get_item(self.MOD_XPATH % "not-installed")

    def test_session_get_item(self):
        with self.conn.start_session("operational") as sess:
            value = sess.get_item(self.MOD_XPATH % "sysrepo-example")
            self.assertIsInstance(value, sysrepo.String)
            self.assertIsInstance(value.value, str)
            self.assertEqual(value.value, "sysrepo-example")

    def test_session_get_items(self):
        with self.conn.start_session("operational") as sess:
            values = sess.get_items(self.MODS_XPATH)
            self.assertIsInstance(values, types.GeneratorType)
            values = list(values)
            self.assertGreater(len(values), 0)

    def test_session_replace_config(self):
        with self.conn.start_session("running") as sess:
            config = {"conf": {"system": {"hostname": "foobar"}}}
            sess.replace_config(config, "sysrepo-example")

    def test_session_copy_config(self):
        with self.conn.start_session("candidate") as sess:
            config = {"conf": {"system": {"hostname": "foobar"}}}
            sess.replace_config(config, "sysrepo-example")
        with self.conn.start_session("running") as sess:
            sess.copy_config("candidate", module_name="sysrepo-example")
            data = sess.get_data("/sysrepo-example:conf")
            self.assertEqual(data, {"conf": {"system": {"hostname": "foobar"}}})

    def test_session_copy_config_errors(self):
        with self.conn.start_session("candidate") as sess:
            config = {"conf": {"system": {"hostname": "foobar"}}}
            sess.replace_config(config, "sysrepo-example")
        with self.conn.start_session("running") as sess:
            with self.assertRaises(ValueError):
                sess.copy_config("invalid")
        with self.conn.start_session("running") as sess:
            with self.assertRaises(sysrepo.SysrepoNotFoundError):
                sess.copy_config("candidate", module_name="not-found")

    def test_session_set_item(self):
        def iface(name, field):
            return "/sysrepo-example:conf/network/interface[name=%r]/%s" % (name, field)

        def assert_data():
            with self.conn.start_session("running") as new_sess:
                data = new_sess.get_data("/sysrepo-example:conf")
                self.assertEqual(
                    data,
                    {
                        "conf": {
                            "network": {
                                "interface": [
                                    {
                                        "name": "eth0",
                                        "address": "1.2.3.4/24",
                                        "up": True,
                                    },
                                    {
                                        "name": "eth1",
                                        "address": "4.3.2.1/24",
                                        "up": False,
                                    },
                                ]
                            }
                        }
                    },
                )

        with self.conn.start_session("running") as sess:
            sess.replace_config({}, "sysrepo-example")
            sess.set_item(iface("eth0", "address"), "1.2.3.4/24")
            sess.set_item(iface("eth0", "up"), True)
            sess.set_item(iface("eth1", "address"), "4.3.2.1/24")
            sess.set_item(iface("eth1", "up"), False)
            sess.apply_changes()
            sess.set_item(iface("eth2", "address"), "8.8.8.8/24")
            sess.set_item(iface("eth2", "up"), True)
            sess.validate()
            assert_data()
            sess.discard_changes()

        assert_data()

    def test_get_netconf_id_and_get_user_are_only_available_in_implicit_session(self):
        with self.conn.start_session("running") as sess:
            with self.assertRaises(sysrepo.SysrepoUnsupportedError):
                sess.get_netconf_id()

            with self.assertRaises(sysrepo.SysrepoUnsupportedError):
                sess.get_user()

    def test_basic_session_lock(self):
        # lock whole datastore
        with self.conn.start_session("running") as sess:
            with sess.locked():
                config = {"conf": {"system": {"hostname": "foobar1"}}}
                sess.replace_config(config, "sysrepo-example")
                sess.apply_changes()

        # lock specific module name with a timeout
        with self.conn.start_session("running") as sess:
            with sess.locked("sysrepo-example", 1000):
                config = {"conf": {"system": {"hostname": "foobar2"}}}
                sess.replace_config(config, "sysrepo-example")
                sess.apply_changes()

    def test_concurrent_session_lock(self):
        def function_thread_one():
            with self.conn.start_session("running") as sess:
                with sess.locked():
                    config = {"conf": {"system": {"hostname": "foobar1"}}}
                    sess.replace_config(config, "sysrepo-example")
                    sess.apply_changes()
                    time.sleep(3)  # keep the lock active

        def function_thread_two():
            time.sleep(1)
            with self.conn.start_session("running") as sess:
                with self.assertRaises(sysrepo.SysrepoLockedError):
                    with sess.locked():
                        pass

        thread_one = threading.Thread(target=function_thread_one)
        thread_two = threading.Thread(target=function_thread_two)

        thread_one.start()
        thread_two.start()

        thread_one.join()
        thread_two.join()

    def test_nacm(self):
        with self.conn.start_session("running") as sess:
            config = {"conf": {"system": {"hostname": "foobar1"}}}
            sess.replace_config(config, "sysrepo-example")
            nacm_config = {
                "ietf-netconf-acm:nacm": {
                    "enable-nacm": True,
                    "read-default": "deny",
                    "groups": {
                        "group": [
                            {"name": "admin", "user-name": ["john"]},
                        ]
                    },
                    "rule-list": [
                        {
                            "name": "sysrepo-example-permit",
                            "group": ["admin"],
                            "rule": [
                                {
                                    "name": "sysrepo-example-permit-read",
                                    "module-name": "sysrepo-example",
                                    "access-operations": "read",
                                    "action": "permit",
                                }
                            ],
                        }
                    ],
                }
            }
            sess.edit_batch(nacm_config, "ietf-netconf-acm", strict=True)
            sess.apply_changes()

        # read access to sysrepo-example is allowed, but write is not for user 'john'
        with self.conn.start_session("running") as sess:
            sess.init_nacm("john")
            data = sess.get_data("/sysrepo-example:conf")
            self.assertEqual(data["conf"]["system"]["hostname"], "foobar1")
            with self.assertRaises(sysrepo.SysrepoUnauthorizedError):
                sess.set_item("/sysrepo-example:conf/system/hostname", "barfoo1")
                sess.apply_changes()



# Copyright (c) 2020 6WIND S.A.
# SPDX-License-Identifier: BSD-3-Clause

from typing import Any, Callable, Optional, Tuple

from _sysrepo import ffi, lib
from .util import c2str


# ------------------------------------------------------------------------------
class SysrepoError(Exception):
    rc = None
    __slots__ = ("msg",)

    def __init__(self, msg: str):
        super().__init__()
        self.msg = msg

    def __str__(self):
        return "%s: %s" % (self.msg, c2str(lib.sr_strerror(self.rc)))

    def __repr__(self):
        return "%s(%r)" % (type(self).__name__, self.msg)

    RC_CLASSES = {}

    @staticmethod
    def register(subclass):
        SysrepoError.RC_CLASSES[subclass.rc] = subclass
        return subclass

    @staticmethod
    def new(msg: str, rc: int) -> "SysrepoError":
        err_class = SysrepoError.RC_CLASSES[rc]
        return err_class(msg)


# ------------------------------------------------------------------------------
@SysrepoError.register
class SysrepoInvalArgError(SysrepoError):
    rc = lib.SR_ERR_INVAL_ARG


@SysrepoError.register
class SysrepoNomemError(SysrepoError):
    rc = lib.SR_ERR_NO_MEMORY


@SysrepoError.register
class SysrepoNotFoundError(SysrepoError):
    rc = lib.SR_ERR_NOT_FOUND


@SysrepoError.register
class SysrepoInternalError(SysrepoError):
    rc = lib.SR_ERR_INTERNAL


@SysrepoError.register
class SysrepoUnsupportedError(SysrepoError):
    rc = lib.SR_ERR_UNSUPPORTED


@SysrepoError.register
class SysrepoValidationFailedError(SysrepoError):
    rc = lib.SR_ERR_VALIDATION_FAILED


@SysrepoError.register
class SysrepoOperationFailedError(SysrepoError):
    rc = lib.SR_ERR_OPERATION_FAILED


@SysrepoError.register
class SysrepoUnauthorizedError(SysrepoError):
    rc = lib.SR_ERR_UNAUTHORIZED


@SysrepoError.register
class SysrepoLockedError(SysrepoError):
    rc = lib.SR_ERR_LOCKED


@SysrepoError.register
class SysrepoTimeOutError(SysrepoError):
    rc = lib.SR_ERR_TIME_OUT


@SysrepoError.register
class SysrepoLyError(SysrepoError):
    rc = lib.SR_ERR_LY


@SysrepoError.register
class SysrepoSysError(SysrepoError):
    rc = lib.SR_ERR_SYS


@SysrepoError.register
class SysrepoExistsError(SysrepoError):
    rc = lib.SR_ERR_EXISTS


@SysrepoError.register
class SysrepoCallbackFailedError(SysrepoError):
    rc = lib.SR_ERR_CALLBACK_FAILED


@SysrepoError.register
class SysrepoCallbackShelveError(SysrepoError):
    rc = lib.SR_ERR_CALLBACK_SHELVE


# ------------------------------------------------------------------------------
def _get_error_msg(session) -> Optional[str]:
    """
    Get the error message information from the given session C pointer.

    :arg "sr_session_ctx_t *" session:
        A session C pointer allocated by libsysrepo.so.
    """
    msg = None
    err_info_p = ffi.new("sr_error_info_t **")
    if lib.sr_session_get_error(session, err_info_p) == lib.SR_ERR_OK:
        err_info = err_info_p[0]
        error_strings = []
        if err_info != ffi.NULL:
            for i in range(err_info.err_count):
                err = err_info.err[i]
                strings = []
                if err.message:
                    strings.append(c2str(err.message))
                if strings:
                    error_strings.append(": ".join(strings))
        msg = ", ".join(error_strings)
    return msg


# ------------------------------------------------------------------------------
def check_call(
    func: Callable[..., int],
    *args: Any,
    valid_codes: Tuple[int, ...] = (lib.SR_ERR_OK,),
) -> int:
    """
    Wrapper around functions of libsysrepo.so.

    :arg func:
        A function from libsysrepo.so that is expected to return an int error
        code.
    :arg args:
        Positional arguments for the function.
    :arg valid_codes:
        Error code values that are considered as a "success". If the function
        returns a value not listed here, a SysrepoError exception will be risen.

    :returns:
        An error code SR_ERR_*.
    :raises SysrepoError:
        If the function returned an error code not listed in valid_codes. If
        the first argument in args is a sr_session_ctx_t object, use it to call
        sr_session_get_error() to get a detailed error message for the risen exception.
    """
    ret = func(*args)
    if ret not in valid_codes:
        msg = None
        if (
            args
            and isinstance(args[0], ffi.CData)
            and ffi.typeof(args[0]) == ffi.typeof("sr_session_ctx_t *")
        ):
            msg = _get_error_msg(args[0])
        if not msg:
            msg = "%s failed" % func.__name__
        raise SysrepoError.new(msg, ret)
    return ret



# Copyright (c) 2020 6WIND S.A.
# SPDX-License-Identifier: BSD-3-Clause

import getpass
import logging
import os
import threading
import time
import typing
import unittest

import sysrepo


YANG_FILE = os.path.join(
    os.path.dirname(os.path.dirname(__file__)), "examples/sysrepo-example.yang"
)
sysrepo.configure_logging(stderr_level=logging.ERROR)


class NotificationSubscriptionTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.conn = sysrepo.SysrepoConnection()
        cls.conn.install_module(YANG_FILE, enabled_features=["turbo"])

    @classmethod
    def tearDownClass(cls):
        # we have to disconnect first to release all resources
        cls.conn.disconnect()
        with sysrepo.SysrepoConnection() as c:
            c.remove_module("sysrepo-example")

    def _test_notification_sub(
        self,
        notif_xpath: str,
        notif_dict: typing.Dict,
        request_extra_info: bool = False,
    ):
        priv = object()
        callback_called = threading.Event()

        def notif_cb(
            xpath, notification_type, notification, timestamp, private_data, **kwargs
        ):
            self.assertEqual(xpath, notif_xpath)
            self.assertEqual(notification_type, "realtime")
            self.assertEqual(notification, notif_dict)
            self.assertIsInstance(timestamp, int)
            self.assertAlmostEqual(timestamp, int(time.time()), delta=5)
            self.assertEqual(private_data, priv)
            if request_extra_info:
                self.assertIn("user", kwargs)
                self.assertEqual(getpass.getuser(), kwargs["user"])
                self.assertIn("netconf_id", kwargs)
                self.assertEqual(kwargs["netconf_id"], 12)
            else:
                self.assertEqual(0, len(kwargs))

            callback_called.set()

        with self.conn.start_session() as listening_session:
            if request_extra_info:
                kwargs = {"extra_info": True}
            else:
                kwargs = {}
            listening_session.subscribe_notification(
                "sysrepo-example", notif_xpath, notif_cb, private_data=priv, **kwargs
            )

            with self.conn.start_session() as sending_session:
                if request_extra_info:
                    sending_session.set_extra_info("netopeer2", 12, getpass.getuser())

                sending_session.notification_send(notif_xpath, notif_dict)
                self.assertTrue(
                    callback_called.wait(timeout=1),
                    "Timed-out while waiting for the notification callback to be called",
                )

    def test_notification_top_level(self):
        self._test_notification_sub(
            notif_xpath="/sysrepo-example:alarm-triggered",
            notif_dict={"description": "An error occurred", "severity": 3},
        )

    def test_notification_nested_in_data_node(self):
        self._test_notification_sub(
            notif_xpath="/sysrepo-example:state/state-changed",
            notif_dict={"message": "Some state changed"},
        )

    def test_notification_sub_with_extra_info(self):
        self._test_notification_sub(
            notif_xpath="/sysrepo-example:state/state-changed",
            notif_dict={"message": "Some state changed"},
            request_extra_info=True,
        )



# Copyright (c) 2020 6WIND S.A.
# SPDX-License-Identifier: BSD-3-Clause

import getpass
import logging
import os
import unittest

import sysrepo


YANG_FILE = os.path.join(
    os.path.dirname(os.path.dirname(__file__)), "examples/sysrepo-example.yang"
)
sysrepo.configure_logging(stderr_level=logging.ERROR)


# ------------------------------------------------------------------------------
class OperSubscriptionTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.conn = sysrepo.SysrepoConnection()
        cls.conn.install_module(YANG_FILE, enabled_features=["turbo"])

    @classmethod
    def tearDownClass(cls):
        # we have to disconnect first to release all resources
        cls.conn.disconnect()
        with sysrepo.SysrepoConnection() as c:
            c.remove_module("sysrepo-example")

    def setUp(self):
        self.sess = self.conn.start_session()

    def tearDown(self):
        self.sess.stop()

    def test_oper_sub(self):
        priv = object()
        state = None

        def oper_data_cb(xpath, private_data):
            self.assertEqual(xpath, "/sysrepo-example:state")
            self.assertEqual(private_data, priv)
            return state

        self.sess.subscribe_oper_data_request(
            "sysrepo-example",
            "/sysrepo-example:state",
            oper_data_cb,
            private_data=priv,
            strict=True,
        )

        with self.conn.start_session("operational") as op_sess:
            state = {
                "state": {
                    "system": {"hostname": "foo"},
                    "network": {
                        "interface": [
                            {
                                "name": "eth0",
                                "address": "1.2.3.4/24",
                                "up": True,
                                "stats": {"rx": 42, "tx": 42},
                            }
                        ]
                    },
                }
            }
            self.assertEqual(op_sess.get_data("/sysrepo-example:state"), state)
            state = {"state": {"invalid": True}}
            with self.assertRaises(sysrepo.SysrepoCallbackFailedError):
                op_sess.get_data("/sysrepo-example:state")

    def test_oper_sub_with_extra_info(self):
        priv = object()
        calls = []

        def oper_data_cb(xpath, private_data, **kwargs):
            self.assertEqual(xpath, "/sysrepo-example:state")
            self.assertEqual(private_data, priv)
            self.assertIn("user", kwargs)
            self.assertEqual(getpass.getuser(), kwargs["user"])
            self.assertIn("netconf_id", kwargs)
            self.assertEqual(kwargs["netconf_id"], 12)
            calls.append((xpath, private_data, kwargs))
            return {"state": {}}

        self.sess.subscribe_oper_data_request(
            "sysrepo-example",
            "/sysrepo-example:state",
            oper_data_cb,
            private_data=priv,
            strict=True,
            extra_info=True,
        )

        with self.conn.start_session("operational") as op_sess:
            op_sess.set_extra_info("netopeer2", 12, getpass.getuser())
            oper_data = op_sess.get_data(
                "/sysrepo-example:state", keep_empty_containers=False
            )
            self.assertEqual(len(calls), 1)
            self.assertEqual(oper_data, {})



# Copyright (c) 2020 6WIND S.A.
# SPDX-License-Identifier: BSD-3-Clause

from contextlib import contextmanager
import logging
import signal
from typing import Dict, Optional, Sequence, Tuple

import libyang

from _sysrepo import ffi, lib
from .errors import SysrepoInternalError, check_call
from .session import SysrepoSession, datastore_value
from .util import c2str, str2c


LOG = logging.getLogger(__name__)


# ------------------------------------------------------------------------------
class SysrepoConnection:
    """
    Create a connection to the sysrepo datastore. If possible (no other connections
    exist), also apply any scheduled changes.

    Do not use `os.fork` after creating a connection. Sysrepo internally stores PID of
    every created connection and this way a mismatch of PID and connection is created.

    The created object can be used as a context manager and will be automatically
    "closed" on exit::

        with sysrepo.SysrepoConnection() as conn:
            # to stuff with conn
        # conn.disconnect() has been called whatever happens
    """

    __slots__ = ("cdata",)

    def __init__(self, cache_running: bool = False, not_printed: bool = True):
        """
        :arg cache_running:
            Always cache running datastore data which makes mainly repeated retrieval of
            data much faster. Affects all sessions created on this connection.
        :arg not_printed:
            The context is created from a sysrepo shared memory and is loaded faster.
            Some operations on the context do not work.
        """
        ctx_flags = 0
        if not_printed:
            ctx_flags |= lib.SR_CTX_NO_PRINTED

        # mandatory flag to work with libyang-python
        ctx_flags |= lib.SR_CTX_SET_PRIV_PARSED

        conn_p = ffi.new("sr_conn_ctx_t **")
        # valid_signals() is only available since python 3.8
        valid_signals = getattr(signal, "valid_signals", lambda: range(1, signal.NSIG))
        sigmask = signal.pthread_sigmask(signal.SIG_BLOCK, valid_signals())
        try:
            lib.sr_cache_running(cache_running)
            lib.sr_context_options(ctx_flags, 0, ffi.NULL)
            check_call(lib.sr_connect, 0, conn_p)
            self.cdata = ffi.gc(conn_p[0], lib.sr_disconnect)

        finally:
            signal.pthread_sigmask(signal.SIG_SETMASK, sigmask)

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        self.disconnect()

    def disconnect(self) -> None:
        """
        Disconnect from the sysrepo datastore.

        Cleans up and frees connection context allocated when instantiating the object.
        All sessions and subscriptions started within the connection will be
        automatically stopped and cleaned up too.

        Connection and all its associated sessions and subscriptions can no longer be
        used even on error.
        """
        if self.cdata is not None:
            if hasattr(ffi, "release"):
                ffi.release(self.cdata)  # causes sr_disconnect to be called
            self.cdata = None

    def start_session(self, datastore: str = "running") -> SysrepoSession:
        """
        Start a new session.

        :arg datastore:
            Datastore on which all sysrepo functions within this session will operate.
            Later on, datastore can be later changed using
            `SysrepoSession.switch_datastore`.

        :returns:
            A `SysrepoSession` object that can be used as a context manager. It will be
            automatically stopped when the manager exits::

                with conn.start_session() as sess:
                    # to stuff with sess
                # sess.stop() has been called whatever happens
        """
        ds = datastore_value(datastore)
        sess_p = ffi.new("sr_session_ctx_t **")
        check_call(lib.sr_session_start, self.cdata, ds, sess_p)
        return SysrepoSession(sess_p[0])

    def acquire_context(self) -> libyang.Context:
        """
        :returns:
            The `libyang.Context` object associated with this connection.
        """
        ctx = lib.sr_acquire_context(self.cdata)
        if not ctx:
            raise SysrepoInternalError("sr_get_context failed")
        return libyang.Context(cdata=ctx)

    def release_context(self):
        lib.sr_release_context(self.cdata)

    @contextmanager
    def get_ly_ctx(self) -> libyang.Context:
        """
        :returns:
            The `libyang.Context` object associated with this connection.
        """
        try:
            yield self.acquire_context()
        finally:
            self.release_context()

    def install_module(
        self,
        filepath: str,
        searchdirs: Optional[str] = None,
        enabled_features: Sequence[str] = (),
        ignore_already_exists=True,
    ) -> None:
        """
        Install a new schema (module) into sysrepo.

        :arg filepath:
            Path to the new schema. Can have either YANG or YIN extension/format.
        :arg searchdirs:
            Optional search directories for import schemas, supports the format
            `<dir>[:<dir>]*`.
        :arg enabled_features:
            Array of enabled features.
        :arg ignore_already_exists:
            Ignore error if module already exists in sysrepo.
        """
        if enabled_features:
            # convert to C strings array
            features = tuple([str2c(f) for f in enabled_features] + [ffi.NULL])
        else:
            features = ffi.NULL

        if ignore_already_exists:
            valid_codes = (lib.SR_ERR_OK, lib.SR_ERR_EXISTS)
        else:
            valid_codes = (lib.SR_ERR_OK,)
        check_call(
            lib.sr_install_module,
            self.cdata,
            str2c(filepath),
            str2c(searchdirs),
            features,
            valid_codes=valid_codes,
        )

    def install_modules(
        self,
        filepaths: Dict[str, Sequence[str]],
        searchdirs: Optional[Sequence[str]] = None,
        ignore_already_exists=False,
    ) -> None:
        """
        Install new schemas (modules) into sysrepo in a batch.

        :arg filepaths:
            Dict of paths to the new schemas associated with a list of features to enable.
            Can have either YANG or YIN extension/format.
        :arg searchdirs:
            Optional list of search directories for import schemas.
        :arg ignore_already_exists:
            Ignore error if module already exists in sysrepo.
        """
        schema_paths = tuple([str2c(k) for k in filepaths] + [ffi.NULL])

        # We need to maintain a pointer to extension names in C
        _ref = []
        all_features = []
        has_feature = False
        for features in filepaths.values():
            features_cname = []
            for f in features:
                cname = str2c(f)
                _ref.append(cname)
                features_cname.append(cname)
                has_feature = True

            all_features.append(ffi.new("char *[]", tuple(features_cname + [ffi.NULL])))

        if has_feature:
            all_features = tuple(all_features)
        else:
            all_features = ffi.NULL

        if ignore_already_exists:
            valid_codes = (lib.SR_ERR_OK, lib.SR_ERR_EXISTS)
        else:
            valid_codes = (lib.SR_ERR_OK,)

        if not searchdirs:
            searchdirs = []

        check_call(
            lib.sr_install_modules,
            self.cdata,
            schema_paths,
            str2c(":".join(searchdirs)),
            all_features,
            valid_codes=valid_codes,
        )

    def update_modules(
        self,
        filepaths: Sequence[str],
        searchdirs: Optional[Sequence[str]] = None,
    ) -> None:
        """
        Update schemas (modules) into sysrepo in a batch.

        :arg filepaths:
            Array of paths to the new schemas.
            Can have either YANG or YIN extension/format.
        :arg searchdirs:
            Optional list of search directories for import schemas.
        """
        schema_paths = tuple([str2c(path) for path in filepaths] + [ffi.NULL])
        if not searchdirs:
            searchdirs = []

        check_call(
            lib.sr_update_modules,
            self.cdata,
            schema_paths,
            str2c(":".join(searchdirs)),
            valid_codes=(lib.SR_ERR_OK,),
        )

    def remove_module(self, name: str, force: bool = False) -> None:
        """
        Remove an installed module from sysrepo. Deferred until there are no
        connections!

        :arg str name:
            Name of the module to remove.
        """
        check_call(lib.sr_remove_module, self.cdata, str2c(name), force)

    def remove_modules(self, names: Sequence[str], force: bool = False) -> None:
        """
        Remove installed modules from sysrepo.

        :arg str names:
            Array of names of the modules to remove.
        """
        names = tuple([str2c(n) for n in names] + [ffi.NULL])
        check_call(lib.sr_remove_modules, self.cdata, names, force)

    def enable_module_feature(self, name: str, feature_name: str) -> None:
        """
        Enable a module feature.

        :arg str name:
            Name of the module.
        :arg str feature_name:
            Name of the feature.
        """
        check_call(
            lib.sr_enable_module_feature, self.cdata, str2c(name), str2c(feature_name)
        )

    def get_module_ds_access(
        self, module_name: str, datastore: str = "running"
    ) -> Tuple[str, str, int]:
        """
        Learn about module permissions.

        :arg str module_name:
            Name of the module.
        :arg str datastore:
            Name of the datastore that will be operated on.
        :returns:
            The owner, group and permissions of the given module name.
            Owner and group are names, not numeric ids.
        """
        owner = ffi.new("char **owner")
        group = ffi.new("char **group")
        perm = ffi.new("mode_t *perm")
        check_call(
            lib.sr_get_module_ds_access,
            self.cdata,
            str2c(module_name),
            datastore_value(datastore),
            owner,
            group,
            perm,
        )
        return (c2str(owner[0]), c2str(group[0]), perm[0])



# Copyright (c) 2020 6WIND S.A.
# SPDX-License-Identifier: BSD-3-Clause

from typing import Any, Optional

from _sysrepo import lib
from .util import c2str


# ------------------------------------------------------------------------------
class Value:
    """
    Abstract class for sysrepo value containing the xpath.
    """

    sr_type = None
    xpath = None
    value_field = None

    def __new__(cls, *args):
        """
        Take 2 arguments by default: ``(python_value, xpath)``.

        If the class holds no value (Container, ContainerPresence & List),
        only take 1 argument: ``xpath``.

        For all cases, xpath is optional. If not specified, it is initialized
        to ``None``.
        """
        if cls is Value:
            raise TypeError("Value cannot be instantiated directly, use subclasses")
        i = 0
        new_args = [cls]
        if cls.value_field is not None:
            try:
                new_args.append(args[i])
                i += 1
            except IndexError:
                raise TypeError("Unspecified value") from None

        self = super().__new__(*new_args)
        try:
            self.xpath = args[i]
        except IndexError:
            pass

        return self

    def __repr__(self) -> str:
        if self.value_field is not None:
            if self.xpath is not None:
                s = "%s: %s" % (self.xpath, self)
            else:
                s = str(self)
        else:
            s = self.xpath or ""
        return "%s(%s)" % (type(self).__name__, s)

    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, type(self))
            and other.xpath == self.xpath
            and other.value == self.value
        )

    def __hash__(self) -> int:
        return hash((type(self), self.xpath, self.value))

    @property
    def value(self) -> Any:
        if self.value_field is not None:
            return type(self).__bases__[-1](self)
        return None

    SR_TYPE_CLASSES = {}

    @staticmethod
    def register(subclass):
        Value.SR_TYPE_CLASSES[subclass.sr_type] = subclass
        return subclass

    @staticmethod
    def parse(cdata) -> Optional["Value"]:
        """
        Parse a 'sr_value_t *' returned by libsysrepo.so and return an instance
        of the correct Value subclass.
        """
        if not cdata:
            return None
        if cdata.type not in Value.SR_TYPE_CLASSES:
            raise TypeError("unknown value type: %r" % cdata.type)
        value_cls = Value.SR_TYPE_CLASSES[cdata.type]
        xpath = c2str(cdata.xpath)
        if value_cls.value_field is not None:
            val = getattr(cdata.data, value_cls.value_field)
            if issubclass(value_cls, str):
                val = c2str(val)
            return value_cls(val, xpath)
        return value_cls(xpath)


# ------------------------------------------------------------------------------
@Value.register
class List(Value):
    sr_type = lib.SR_LIST_T


@Value.register
class Container(Value):
    sr_type = lib.SR_CONTAINER_T


@Value.register
class ContainerPresence(Value):
    sr_type = lib.SR_CONTAINER_PRESENCE_T


@Value.register
class LeafEmpty(Value):
    sr_type = lib.SR_LEAF_EMPTY_T


@Value.register
class Int8(Value, int):
    sr_type = lib.SR_INT8_T
    value_field = "int8_val"


@Value.register
class Int16(Value, int):
    sr_type = lib.SR_INT16_T
    value_field = "int16_val"


@Value.register
class Int32(Value, int):
    sr_type = lib.SR_INT32_T
    value_field = "int32_val"


@Value.register
class Int64(Value, int):
    sr_type = lib.SR_INT64_T
    value_field = "int64_val"


@Value.register
class UInt8(Value, int):
    sr_type = lib.SR_UINT8_T
    value_field = "uint8_val"


@Value.register
class UInt16(Value, int):
    sr_type = lib.SR_UINT16_T
    value_field = "uint16_val"


@Value.register
class UInt32(Value, int):
    sr_type = lib.SR_UINT32_T
    value_field = "uint32_val"


@Value.register
class UInt64(Value, int):
    sr_type = lib.SR_UINT64_T
    value_field = "uint64_val"


@Value.register
class String(Value, str):
    sr_type = lib.SR_STRING_T
    value_field = "string_val"


@Value.register
class Bits(Value, str):
    sr_type = lib.SR_BITS_T
    value_field = "bits_val"


@Value.register
class Enum(Value, str):
    sr_type = lib.SR_ENUM_T
    value_field = "enum_val"


@Value.register
class Binary(Value, str):
    sr_type = lib.SR_BINARY_T
    value_field = "binary_val"


@Value.register
class AnyXML(Value, str):
    sr_type = lib.SR_ANYXML_T
    value_field = "anyxml_val"


@Value.register
class AnyData(Value, str):
    sr_type = lib.SR_ANYDATA_T
    value_field = "anydata_val"


@Value.register
class IdentityRef(Value, str):
    sr_type = lib.SR_IDENTITYREF_T
    value_field = "identityref_val"


@Value.register
class InstanceId(Value, str):
    sr_type = lib.SR_INSTANCEID_T
    value_field = "instanceid_val"


@Value.register
class Decimal64(Value, float):
    sr_type = lib.SR_DECIMAL64_T
    value_field = "decimal64_val"


@Value.register
class Bool(Value, int):  # bool cannot be subclassed...
    sr_type = lib.SR_BOOL_T
    value_field = "bool_val"

    @property
    def value(self) -> bool:
        return bool(self)

    def __str__(self) -> str:
        return "True" if self else "False"



# Copyright (c) 2020 6WIND S.A.
# SPDX-License-Identifier: BSD-3-Clause

import functools
import inspect
import logging
from typing import Any, Optional

from _sysrepo import ffi, lib


# ------------------------------------------------------------------------------
def str2c(s: Optional[str]):
    if s is None:
        return ffi.NULL
    if hasattr(s, "encode"):
        s = s.encode("utf-8")
    return ffi.new("char []", s)


# ------------------------------------------------------------------------------
def c2str(c) -> Optional[str]:
    if c == ffi.NULL:
        return None
    s = ffi.string(c)
    if hasattr(s, "decode"):
        s = s.decode("utf-8")
    return s


# ------------------------------------------------------------------------------
def is_async_func(func: Any) -> bool:
    if inspect.iscoroutinefunction(func):
        return True
    if isinstance(func, functools.partial):
        return is_async_func(func.func)
    return False


# ------------------------------------------------------------------------------
LOG = logging.getLogger("sysrepo")
LOG.addHandler(logging.NullHandler())
LOG_LEVELS_SR2PY = {
    lib.SR_LL_ERR: logging.ERROR,
    lib.SR_LL_WRN: logging.WARNING,
    lib.SR_LL_INF: logging.INFO,
    lib.SR_LL_DBG: logging.DEBUG,
}
LOG_LEVELS_PY2SR = {v: k for k, v in LOG_LEVELS_SR2PY.items()}
LOG_LEVELS_PY2SR[logging.CRITICAL] = lib.SR_LL_ERR


@ffi.def_extern(name="srpy_log_cb")
def log_callback(level, msg):
    py_level = LOG_LEVELS_SR2PY.get(level, logging.NOTSET)
    LOG.log(py_level, "%s", c2str(msg))


# ------------------------------------------------------------------------------
def configure_logging(
    *,
    stderr_level: int = logging.NOTSET,
    syslog_level: int = logging.NOTSET,
    syslog_app_name: str = "sysrepo",
    py_logging: bool = False,
):
    """
    Configure logging for sysrepo. By default, all logging is disabled.

    All *_level arguments take standard python logging levels.

    :arg stderr_level:
        The level for standard error logging. If set to logging.NOTSET, standard error
        logging is disabled (the default).
    :arg syslog_level:
        The level for syslog logging. If set to logging.NOTSET, syslog logging is
        disabled (the default).
    :arg syslog_app_name:
        The name of the application sent in syslog messages.
    :arg py_logging:
        If True, enable logging via the python logging system. Logging level is
        controlled by python logging system.
    """
    sr_stderr_level = LOG_LEVELS_PY2SR.get(stderr_level, lib.SR_LL_NONE)
    lib.sr_log_stderr(sr_stderr_level)
    sr_syslog_level = LOG_LEVELS_PY2SR.get(syslog_level, lib.SR_LL_NONE)
    lib.sr_log_syslog(str2c(syslog_app_name), sr_syslog_level)
    if py_logging:
        lib.sr_log_set_cb(lib.srpy_log_cb)
    else:
        lib.sr_log_set_cb(ffi.NULL)


# ------------------------------------------------------------------------------
def get_syslog_level() -> int:
    """
    Return current syslog log level.

    :returns int:
        The log level.
    """
    return LOG_LEVELS_SR2PY.get(lib.sr_log_get_syslog(), logging.NOTSET)


# ------------------------------------------------------------------------------
def get_stderr_level() -> int:
    """
    Return current stderr log level.

    :returns int:
        The log level.
    """
    return LOG_LEVELS_SR2PY.get(lib.sr_log_get_stderr(), logging.NOTSET)



# Copyright (c) 2020 6WIND S.A.
# SPDX-License-Identifier: BSD-3-Clause

import getpass
import logging
import os
import unittest

import libyang

import sysrepo


YANG_FILE = os.path.join(
    os.path.dirname(os.path.dirname(__file__)), "examples/sysrepo-example.yang"
)
sysrepo.configure_logging(stderr_level=logging.ERROR)


# ------------------------------------------------------------------------------
class RpcSubscriptionTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.conn = sysrepo.SysrepoConnection()
        cls.conn.install_module(YANG_FILE, enabled_features=["turbo"])

    @classmethod
    def tearDownClass(cls):
        # we have to disconnect first to release all resources
        cls.conn.disconnect()
        with sysrepo.SysrepoConnection() as c:
            c.remove_module("sysrepo-example")

    def test_rpc_sub(self):
        priv = object()
        calls = []
        rpc_xpath = "/sysrepo-example:poweroff"

        def rpc_cb(xpath, input_params, event, private_data):
            calls.append((xpath, input_params, event, private_data))
            behaviour = input_params.get("behaviour")
            if behaviour == "failure":
                raise sysrepo.SysrepoCallbackFailedError("poweroff failed")
            if behaviour == "bad-output":
                return {"unkown-leaf": False}
            return {"message": "bye bye"}

        with self.conn.start_session() as sess:
            sess.subscribe_rpc_call(rpc_xpath, rpc_cb, private_data=priv, strict=True)

            with self.conn.start_session() as rpc_sess:
                input_params = {"behaviour": "failure"}
                with self.assertRaises(sysrepo.SysrepoCallbackFailedError):
                    rpc_sess.rpc_send(rpc_xpath, input_params)
                self.assertEqual(len(calls), 1)
                self.assertEqual(calls[0], (rpc_xpath, input_params, "rpc", priv))
                del calls[:]

                input_params = {"behaviour": "bad-output"}
                with self.assertRaises(sysrepo.SysrepoCallbackFailedError):
                    rpc_sess.rpc_send(rpc_xpath, input_params)
                self.assertEqual(len(calls), 1)
                self.assertEqual(calls[0], (rpc_xpath, input_params, "rpc", priv))
                del calls[:]

                input_params = {"behaviour": "success"}
                output = rpc_sess.rpc_send(rpc_xpath, input_params)
                self.assertEqual(output, {"message": "bye bye"})
                self.assertEqual(len(calls), 1)
                self.assertEqual(calls[0], (rpc_xpath, input_params, "rpc", priv))
                del calls[:]

    def test_action_sub(self):
        priv = object()
        calls = []
        subs_xpath = "/sysrepo-example:conf/security/alarm/trigger"

        def action_cb(xpath, input_params, event, private_data):
            calls.append((xpath, input_params, event, private_data))
            _, _, keys = list(libyang.xpath_split(xpath))[2]
            _, name = keys[0]
            duration = input_params["duration"]
            return {
                "message": "%s alarm triggered for %s seconds"
                % (
                    name,
                    duration,
                )
            }

        def module_change_cb(event, req_id, changes, private_data):
            # unused
            pass

        with self.conn.start_session() as sess:
            sess.subscribe_rpc_call(
                subs_xpath, action_cb, private_data=priv, strict=True
            )
            sess.replace_config(
                {
                    "conf": {
                        "security": {
                            "alarm": [
                                {"name": "lab1"},
                                {"name": "lab2"},
                                {"name": "office1"},
                            ],
                        },
                    },
                },
                "sysrepo-example",
            )
            # Subscribe to data changes so that config is present in operational
            # datastore. This is required for action input validation.
            sess.subscribe_module_change("sysrepo-example", None, module_change_cb)

            with self.conn.start_session() as rpc_sess:
                xpath = "/sysrepo-example:conf/security/alarm[name='lab2']/trigger"
                output = rpc_sess.rpc_send(xpath, {"duration": 30})
                self.assertEqual(
                    output, {"message": "lab2 alarm triggered for 30 seconds"}
                )
                self.assertEqual(len(calls), 1)
                self.assertEqual(calls[0], (xpath, {"duration": 30}, "rpc", priv))
                del calls[:]

                # no value for duration, check default value is set
                xpath = "/sysrepo-example:conf/security/alarm[name='office1']/trigger"
                output = rpc_sess.rpc_send(xpath, {})
                self.assertEqual(
                    output, {"message": "office1 alarm triggered for 1 seconds"}
                )
                self.assertEqual(len(calls), 1)
                self.assertEqual(calls[0], (xpath, {"duration": 1}, "rpc", priv))
                del calls[:]

    def test_rpc_sub_with_extra_info(self):
        priv = object()
        calls = []
        rpc_xpath = "/sysrepo-example:poweroff"

        def rpc_cb(xpath, input_params, event, private_data, **kwargs):
            self.assertEqual(rpc_xpath, xpath)
            self.assertEqual(input_params, {"behaviour": "success"})
            self.assertEqual(event, "rpc")
            self.assertIs(private_data, priv)
            self.assertIn("user", kwargs)
            self.assertEqual(getpass.getuser(), kwargs["user"])
            self.assertIn("netconf_id", kwargs)
            self.assertEqual(kwargs["netconf_id"], 12)
            calls.append((xpath, input_params, event, private_data))
            return {"message": "bye bye"}

        with self.conn.start_session() as sess:
            sess.subscribe_rpc_call(
                rpc_xpath, rpc_cb, private_data=priv, strict=True, extra_info=True
            )

            with self.conn.start_session() as rpc_sess:
                rpc_sess.set_extra_info("netopeer2", 12, getpass.getuser())
                output = rpc_sess.rpc_send(rpc_xpath, {"behaviour": "success"})
                self.assertEqual(len(calls), 1)
                self.assertEqual(output, {"message": "bye bye"})



# Copyright (c) 2020 6WIND S.A.
# SPDX-License-Identifier: BSD-3-Clause

from contextlib import contextmanager
import inspect
import logging
from typing import Any, Callable, Dict, Iterator, List, Optional

import libyang

from _sysrepo import ffi, lib
from .change import Change
from .errors import (
    SysrepoError,
    SysrepoInternalError,
    SysrepoNotFoundError,
    SysrepoUnsupportedError,
    check_call,
)
from .subscription import Subscription
from .util import c2str, is_async_func, str2c
from .value import Value


LOG = logging.getLogger(__name__)


# ------------------------------------------------------------------------------
class SysrepoSession:
    """
    Python representation of `sr_session_ctx_t *`.

    .. attention::

        Do not instantiate this class manually, use `SysrepoConnection.start_session`.
    """

    __slots__ = (
        "cdata",
        "is_implicit",
        "subscriptions",
        "nacm_subscription",
    )

    # begin: general
    def __init__(self, cdata, implicit: bool = False):
        """
        :arg "sr_session_ctx_t *" cdata:
            The session pointer allocated by :func:`SysrepoConnection.start_session`.
        :arg implicit:
            Used to identify sessions provided in subscription callbacks.
        """
        self.cdata = cdata
        self.is_implicit = implicit
        self.subscriptions = []
        self.nacm_subscription = None

    def __enter__(self) -> "SysrepoSession":
        return self

    def init_nacm(self, nacm_user: str, no_thread: bool = False) -> None:
        """
        Set the NACM user for this session. This is used to determine the effective user
        for NACM checks.

        :arg nacm_user:
            The NACM user name to be set.
        :arg no_thread:
            If True, no thread will be created for handling NACM subscription meaning.
            Default to False.
        """
        flags = _subscribe_flags(
            no_thread=no_thread,
        )
        if self.is_implicit:
            raise SysrepoUnsupportedError("cannot set NACM for implicit sessions")

        sub_p = ffi.new("sr_subscription_ctx_t **")
        check_call(lib.sr_nacm_init, self.cdata, flags, sub_p)
        check_call(lib.sr_nacm_set_user, self.cdata, str2c(nacm_user))
        self.nacm_subscription = sub_p

    def __exit__(self, *args, **kwargs) -> None:
        self.stop()

    def stop(self) -> None:
        """
        Stop current session and releases resources tied to the session.

        Also releases any locks held and frees subscriptions created (only) by this
        session.
        """
        if self.cdata is None:
            return  # already stopped
        if self.is_implicit:
            raise SysrepoUnsupportedError("implicit sessions cannot be stopped")
        if self.nacm_subscription is not None:
            check_call(lib.sr_unsubscribe, self.nacm_subscription[0])
            self.nacm_subscription = None
            lib.sr_nacm_destroy()

        # clear subscriptions
        while self.subscriptions:
            sub = self.subscriptions.pop()
            try:
                sub.unsubscribe()
            except Exception:
                LOG.exception("Subscription.unsubscribe failed")

        # stop session
        try:
            check_call(lib.sr_session_stop, self.cdata)
        finally:
            self.cdata = None

    def release_context(self):
        conn = lib.sr_session_get_connection(self.cdata)
        lib.sr_release_context(conn)

    def acquire_context(self) -> libyang.Context:
        """
        :returns:
            The libyang context object associated with this session.
        """
        conn = lib.sr_session_get_connection(self.cdata)
        if not conn:
            raise SysrepoInternalError("sr_session_get_connection failed")
        ctx = lib.sr_acquire_context(conn)
        if not ctx:
            raise SysrepoInternalError("sr_get_context failed")

        return libyang.Context(cdata=ctx)

    def get_datastore(self) -> str:
        """
        Get the datastore a session operates on.

        :returns str:
            The datastore name.
        """
        return datastore_name(lib.sr_session_get_ds(self.cdata))

    def lock(self, module_name: str = "", timeout_ms: int = 0) -> None:
        """
        Locks the data of the specified module or the whole datastore.
        :arg module_name:
            Optional name of the module to be locked.
        :arg timeout_ms:
            Optional timeout in ms for waiting. If 0, no waiting is performed.
        """
        module = str2c(module_name) if len(module_name) > 0 else ffi.NULL
        check_call(lib.sr_lock, self.cdata, module, timeout_ms)

    def unlock(self, module_name: str = "") -> None:
        """
        Unlocks the data of the specified module or the whole datastore.
        :arg module_name:
            Optional name of the module to be locked.
        """
        module = str2c(module_name) if len(module_name) > 0 else ffi.NULL
        check_call(lib.sr_unlock, self.cdata, module)

    @contextmanager
    def locked(self, module_name: str = "", timeout_ms: int = 0):
        """
        Convenience method which manages un-/locking the data of the specified module
        or the whole datastore.
        :arg module_name:
            Optional name of the module to be locked.
        :arg timeout_ms:
            Optional timeout in ms for waiting. If 0, no waiting is performed.
        """
        try:
            self.lock(module_name, timeout_ms)
            yield
        finally:
            self.unlock(module_name)

    def switch_datastore(self, datastore: str) -> None:
        """
        Change datastore which the session operates on. All subsequent calls will be
        issued on the chosen datastore. Previous calls are not affected (previous
        subscriptions, for instance).

        :arg str datastore:
            New datastore that will be operated on. Can be one of `running`,
            `startup`, `operational` or `candidate`.
        """
        if self.is_implicit:
            raise SysrepoUnsupportedError(
                "cannot change datastore of implicit sessions"
            )
        ds = datastore_value(datastore)
        check_call(lib.sr_session_switch_ds, self.cdata, ds)

    def set_error(self, err: BaseException):
        """
        Set detailed error information into provided session. Used to notify the client
        library about errors that occurred in the application code. Does not print the
        message.

        Intended for change, RPC/action, or operational callbacks to be used on the
        provided session.

        :arg BaseException err:
            The python exception object that carries the error information.
        """
        if not self.is_implicit:
            raise SysrepoUnsupportedError("can only report errors on implicit sessions")
        if isinstance(err, SysrepoError):
            code = err.rc
            message = err.msg
        else:
            code = lib.SR_ERR_CALLBACK_FAILED
            message = str(err)
        check_call(
            lib.sr_session_set_error,
            self.cdata,
            ffi.NULL,
            code,
            str2c("%s"),
            str2c(message),
        )

    def get_originator_name(self) -> str:
        """
        It can only be called on an implicit sysrepo.Session (i.e., it can only be
        called from an event callback).

        :returns: the originator name for the event originator sysrepo session
        """
        if not self.is_implicit:
            raise SysrepoUnsupportedError(
                "can only report originator name on implicit sessions"
            )

        return c2str(lib.sr_session_get_orig_name(self.cdata))

    def set_extra_info(self, originator_name: str, netconf_id: int, user: str) -> None:
        """
        Set the Session extra infos.
        This function is mainly used for testing purpose.

        :arg str originator_name:
            The originator name, should be netopeer2.
        :arg int netconf_id:
            The netconf_id.
        :arg str originator_name:
            The user name.
        """
        # orig name
        check_call(lib.sr_session_set_orig_name, self.cdata, str2c(originator_name))

        # netconf_id
        c_netconf_id = ffi.new("uint32_t *")
        c_netconf_id[0] = netconf_id
        p_netconf_id = ffi.cast("const void **", c_netconf_id)
        check_call(
            lib.sr_session_push_orig_data,
            self.cdata,
            ffi.sizeof(c_netconf_id),
            p_netconf_id,
        )

        # user
        c_user = ffi.new("char[]", user.encode())
        p_user = ffi.cast("const void **", c_user)
        check_call(
            lib.sr_session_push_orig_data, self.cdata, ffi.sizeof(c_user), p_user
        )

    def get_netconf_id(self) -> int:
        """
        It can only be called on an implicit sysrepo.Session (i.e., it can only be
        called from an event callback).

        :returns: the NETCONF session ID set for the event originator sysrepo session
        """
        if not self.is_implicit:
            raise SysrepoUnsupportedError(
                "can only report netconf id on implicit sessions"
            )

        if self.get_originator_name() != "netopeer2":
            raise SysrepoUnsupportedError("can only report netconf id for netopeer2")

        size = ffi.new("uint32_t *")
        p_nc_id = ffi.new("const void **")

        check_call(lib.sr_session_get_orig_data, self.cdata, 0, size, p_nc_id)

        nc_id = ffi.cast("uint32_t *", p_nc_id[0])
        return nc_id[0]

    def get_user(self) -> str:
        """
        It can only be called on an implicit sysrepo.Session (i.e., it can only be
        called from an event callback)

        :returns: the effective username of the event originator sysrepo session
        """
        if not self.is_implicit:
            raise SysrepoUnsupportedError("can only report user on implicit sessions")

        size = ffi.new("uint32_t *")
        p_user = ffi.new("const void **")

        check_call(lib.sr_session_get_orig_data, self.cdata, 1, size, p_user)

        user = ffi.cast("const char *", p_user[0])
        return c2str(user)

    @contextmanager
    def get_ly_ctx(self) -> libyang.Context:
        """
        :returns:
            The libyang context object associated with this session.
        """
        try:
            yield self.acquire_context()
        finally:
            self.release_context()

    # end: general

    # begin: subscription
    ModuleChangeCallbackType = Callable[[str, int, List[Change], Any], None]
    """
    Callback to be called when the change in the datastore occurs.

    :arg event:
        Type of the callback event that has occurred. Can be one of: "update", "change",
        "done", "abort", "enabled".
    :arg req_id:
        Request ID unique for the specific module name. Connected events for one request
        ("change" and "done" for example) have the same request ID.
    :arg changes:
        List of `sysrepo.Change` objects representing what parts of the configuration
        have changed.
    :arg private_data:
        Private context opaque to sysrepo used when subscribing.
    :arg kwargs (optional):
        If the callback was registered with the argument extra_info=True (see
        Session.subscribe_module_change), then extra keyword arguments are passed when
        calling the callback:
            * netconf_id: the NETCONF session ID set for the event originator
                sysrepo session
            * user: the effective username of the event originator sysrepo session

    When event is one of ("update", "change"), if the callback raises an exception, the
    changes will be rejected and the error will be forwarded to the client that made the
    change. If the exception is a subclass of `SysrepoError`, the traceback will not be
    sent to the logging system. For consistency and to avoid confusion with unexpected
    errors, the callback should raise explicit `SysrepoValidationFailedError` exceptions
    to reject changes.
    """

    # pylint: disable=too-many-arguments
    def subscribe_module_change(
        self,
        module: str,
        xpath: Optional[str],
        callback: ModuleChangeCallbackType,
        *,
        priority: int = 0,
        no_thread: bool = False,
        passive: bool = False,
        done_only: bool = False,
        enabled: bool = False,
        update: bool = False,
        filter_origin: bool = False,
        private_data: Any = None,
        asyncio_register: bool = False,
        include_implicit_defaults: bool = True,
        include_deleted_values: bool = False,
        extra_info: bool = False,
    ) -> None:
        """
        Subscribe for changes made in the specified module.

        :arg module:
            Name of the module of interest for change notifications.
        :arg xpath:
            Optional xpath further filtering the changes that will be handled
            by this subscription.
        :arg callback:
            Callback to be called when the change in the datastore occurs.
        :arg priority:
            Specifies the order in which the callbacks (**within module**) will
            be called.
        :arg no_thread:
            There will be no thread created for handling this subscription
            meaning no event will be processed! Default to `True` if
            asyncio_register is `True`.
        :arg passive:
            The subscriber is not the "owner" of the subscribed data tree, just
            a passive watcher for changes.
        :arg done_only:
            The subscriber does not support verification of the changes and
            wants to be notified only after the changes has been applied in the
            datastore, without the possibility to deny them.
        :arg enabled:
            The subscriber wants to be notified about the current configuration
            at the moment of subscribing.
        :arg update:
            The subscriber wants to be called before the configuration is applied.
        :arg filter_origin:
            Filter events on the originator side to unburden the subscriber, but
            results in 0 value for filtered-out changes in the subscriber infos.
        :arg private_data:
            Private context passed to the callback function, opaque to sysrepo.
        :arg asyncio_register:
            Add the created subscription event pipe into asyncio event loop
            monitored read file descriptors. Implies `no_thread=True`.
        :arg include_implicit_defaults:
            Include implicit default nodes in changes.
        :arg include_deleted_values:
            Include deleted nodes values in changes.
        :arg extra_info:
            When True, the given callback is called with extra keyword arguments
            containing extra information of the sysrepo session that gave origin to the
            event (see ModuleChangeCallbackType for more details)
        """
        if self.is_implicit:
            raise SysrepoUnsupportedError("cannot subscribe with implicit sessions")
        _check_subscription_callback(callback, self.ModuleChangeCallbackType)

        sub = Subscription(
            callback,
            private_data,
            asyncio_register=asyncio_register,
            include_implicit_defaults=include_implicit_defaults,
            include_deleted_values=include_deleted_values,
            extra_info=extra_info,
        )
        sub_p = ffi.new("sr_subscription_ctx_t **")

        if asyncio_register:
            no_thread = True  # we manage our own event loop
        flags = _subscribe_flags(
            no_thread=no_thread,
            passive=passive,
            done_only=done_only,
            enabled=enabled,
            update=update,
            filter_origin=filter_origin,
        )

        check_call(
            lib.sr_module_change_subscribe,
            self.cdata,
            str2c(module),
            str2c(xpath),
            lib.srpy_module_change_cb,
            sub.handle,
            priority,
            flags,
            sub_p,
        )
        sub.init(sub_p[0])

        self.subscriptions.append(sub)

    UnsafeModuleChangeCallbackType = Callable[["SysrepoSession", str, int, Any], None]
    """
    Callback to be called when the change in the datastore occurs. Provides implicit 
    session object instead of list of changes. THE CALLBACK SHOULD NEVER KEEP A
    REFERENCE ON THE IMPLICIT SESSION OBJECT TO AVOID USER-AFTER-FREE BUGS.

    :arg session:
        Implicit session (do not stop) with information about the changed data.
    :arg event:
        Type of the callback event that has occurred. Can be one of: "update", "change",
        "done", "abort", "enabled".
    :arg req_id:
        Request ID unique for the specific module name. Connected events for one request
        ("change" and "done" for example) have the same request ID.
    :arg private_data:
        Private context opaque to sysrepo used when subscribing.

    When event is one of ("update", "change"), if the callback raises an exception, the
    changes will be rejected and the error will be forwarded to the client that made the
    change. If the exception is a subclass of `SysrepoError`, the traceback will not be
    sent to the logging system. For consistency and to avoid confusion with unexpected
    errors, the callback should raise explicit `SysrepoValidationFailedError` exceptions
    to reject changes.
    """

    def subscribe_module_change_unsafe(
        self,
        module: str,
        xpath: Optional[str],
        callback: UnsafeModuleChangeCallbackType,
        *,
        priority: int = 0,
        no_thread: bool = False,
        passive: bool = False,
        done_only: bool = False,
        enabled: bool = False,
        update: bool = False,
        filter_origin: bool = False,
        private_data: Any = None,
        asyncio_register: bool = False,
    ) -> None:
        """
        Subscribe for changes made in the specified module. Implicit session object
        is returned to callback instead of list of changes. When callback returns to
        Sysrepo, the session is freed, so we cannot keep a reference on it and
        schedule async code to run later. For this reason, async callbacks are NOT
        allowed here.

        :arg module:
            Name of the module of interest for change notifications.
        :arg xpath:
            Optional xpath further filtering the changes that will be handled
            by this subscription.
        :arg callback:
            Callback to be called when the change in the datastore occurs.
        :arg priority:
            Specifies the order in which the callbacks (**within module**) will
            be called.
        :arg no_thread:
            There will be no thread created for handling this subscription
            meaning no event will be processed! Default to `True` if
            asyncio_register is `True`.
        :arg passive:
            The subscriber is not the "owner" of the subscribed data tree, just
            a passive watcher for changes.
        :arg done_only:
            The subscriber does not support verification of the changes and
            wants to be notified only after the changes has been applied in the
            datastore, without the possibility to deny them.
        :arg enabled:
            The subscriber wants to be notified about the current configuration
            at the moment of subscribing.
        :arg update:
            The subscriber wants to be called before the configuration is applied.
        :arg filter_origin:
            Filter events on the originator side to unburden the subscriber, but
            results in 0 value for filtered-out changes in the subscriber infos.
        :arg private_data:
            Private context passed to the callback function, opaque to sysrepo.
        :arg asyncio_register:
            Add the created subscription event pipe into asyncio event loop
            monitored read file descriptors. Implies `no_thread=True`.
        """
        if self.is_implicit:
            raise SysrepoUnsupportedError("cannot subscribe with implicit sessions")
        if is_async_func(callback):
            raise SysrepoUnsupportedError(
                "cannot use unsafe subscription with async callback"
            )
        _check_subscription_callback(callback, self.UnsafeModuleChangeCallbackType)

        sub = Subscription(
            callback,
            private_data,
            asyncio_register=asyncio_register,
            unsafe=True,
        )
        sub_p = ffi.new("sr_subscription_ctx_t **")

        if asyncio_register:
            no_thread = True  # we manage our own event loop
        flags = _subscribe_flags(
            no_thread=no_thread,
            passive=passive,
            done_only=done_only,
            enabled=enabled,
            update=update,
            filter_origin=filter_origin,
        )
        check_call(
            lib.sr_module_change_subscribe,
            self.cdata,
            str2c(module),
            str2c(xpath),
            lib.srpy_module_change_cb,
            sub.handle,
            priority,
            flags,
            sub_p,
        )
        sub.init(sub_p[0])

        self.subscriptions.append(sub)

    OperDataCallbackType = Callable[[str, Any], Optional[Dict]]
    """
    Callback to be called when the operational data are requested.

    :arg xpath:
        The XPath requested by a client. Can be None if the client requested for all the
        module operational data.
    :arg private_data:
        Private context opaque to sysrepo used when subscribing.
    :arg kwargs (optional):
        If the callback was registered with the argument extra_info=True (see
        Session.subscribe_module_change), then extra keyword arguments are passed when
        calling the callback:
            * netconf_id: the NETCONF session ID set for the event originator
                sysrepo session
            * user: the effective username of the event originator sysrepo session

    The callback is expected to return a python dictionary containing the operational
    data. The dictionary should be in the libyang "dict" format. It will be parsed to a
    libyang lyd_node before returning to sysrepo using `libyang.Module.parse_data_dict`.
    If the callback returns `None`, nothing is returned to sysrepo. If the callback
    raises an exception, the error message is forwarded to the client that requested for
    data. If the exception is a subclass of `SysrepoError`, no traceback is sent to the
    logging system.
    """

    def subscribe_oper_data_request(
        self,
        module: str,
        xpath: str,
        callback: OperDataCallbackType,
        *,
        no_thread: bool = False,
        oper_merge: bool = False,
        private_data: Any = None,
        asyncio_register: bool = False,
        strict: bool = False,
        extra_info: bool = False,
    ) -> None:
        """
        Register for providing operational data at the given xpath.

        :arg module:
            Name of the affected module.
        :arg xpath:
            Xpath identifying the subtree which the provider is able to provide.
            Predicates can be used to provide only specific instances of nodes.
        :arg callback:
            Callback to be called when the operational data for the given xpath are
            requested.
        :arg no_thread:
            There will be no thread created for handling this subscription meaning no
            event will be processed! Default to `True` if asyncio_register is `True`.
        :arg oper_merge:
            Instead of removing any previous existing matching data before
            getting them from an operational subscription callback, keep
            them. Then the returned data are merged into the existing
            data. Valid only for operational subscriptions.
        :arg private_data:
            Private context passed to the callback function, opaque to sysrepo.
        :arg asyncio_register:
            Add the created subscription event pipe into asyncio event loop monitored
            read file descriptors. Implies `no_thread=True`.
        :arg strict:
            Reject the whole data returned by callback if it contains elements without
            schema definition.
        :arg extra_info:
            When True, the given callback is called with extra keyword arguments
            containing extra information of the sysrepo session that gave origin to the
            event (see OperDataCallbackType for more details)
        """
        if self.is_implicit:
            raise SysrepoUnsupportedError("cannot subscribe with implicit sessions")
        _check_subscription_callback(callback, self.OperDataCallbackType)

        sub = Subscription(
            callback,
            private_data,
            asyncio_register=asyncio_register,
            strict=strict,
            extra_info=extra_info,
        )
        sub_p = ffi.new("sr_subscription_ctx_t **")

        if asyncio_register:
            no_thread = True  # we manage our own event loop
        flags = _subscribe_flags(no_thread=no_thread, oper_merge=oper_merge)

        check_call(
            lib.sr_oper_get_subscribe,
            self.cdata,
            str2c(module),
            str2c(xpath),
            lib.srpy_oper_data_cb,
            sub.handle,
            flags,
            sub_p,
        )
        sub.init(sub_p[0])

        self.subscriptions.append(sub)

    RpcCallbackType = Callable[[str, Dict, str, Any], Optional[Dict]]
    """
    Callback to be called when the RPC/action is invoked.

    :arg xpath:
        The full data path to the invoked RPC/action. When it is an RPC, the form is
        `/prefix:rpc-name`. When it is an action, it is the full data path with all
        parent nodes: `/prefix:container/list[key="val"]/action-name`.
    :arg input_params:
        The input arguments in a python dictionary. The contents are limited to the
        children of the "input" node. For example, with a YANG rpc defined like this::

            rpc rpc-name {
              input {
                leaf param1 {
                  type uint32;
                }
                leaf param2 {
                  type string;
                }
              }
              output {
                leaf foo {
                  type int8;
                }
                leaf bar {
                  type string;
                }
              }
            }

        The input_params dict may look like this::

            {'param1': 42, 'param2': 'foobar'}

        If there are no input parameters provided by the client, the dict will be empty.

        For actions, the xpath argument allows to determine the parent node of the
        action input parameters.
    :arg event:
        In most cases, it is always 'rpc'. When multiple callbacks are registered for
        the same RPC and one of the callbacks failed. The remainder of the callbacks
        will be called with 'abort'.
    :arg private_data:
        Private context opaque to sysrepo used when subscribing.
    :arg kwargs (optional):
        If the callback was registered with the argument extra_info=True (see
        Session.subscribe_module_change), then extra keyword arguments are passed when
        calling the callback:
            * netconf_id: the NETCONF session ID set for the event originator
                sysrepo session
            * user: the effective username of the event originator sysrepo session

    The callback is expected to return a python dictionary containing the RPC output
    data. The dictionary should be in the libyang "dict" format and must only contain
    the actual output parameters without any parents. Using the previous example::

        {'foo': 47, 'bar': 'baz'}

    It will be passed to libyang.DRpc.merge_data_dict() to return output data to
    sysrepo. If the callback returns None, nothing is returned to sysrepo. If the
    callback raises an exception, the error message is forwarded to the client that
    called the RPC. If the exception is a subclass of SysrepoError, no traceback is sent
    to the logging system.
    """

    def subscribe_rpc_call(
        self,
        xpath: str,
        callback: RpcCallbackType,
        *,
        priority: int = 0,
        no_thread: bool = False,
        private_data: Any = None,
        asyncio_register: bool = False,
        strict: bool = False,
        include_implicit_defaults: bool = True,
        extra_info: bool = False,
    ) -> None:
        """
        Subscribe for the delivery of an RPC/action.

        :arg xpath:
            XPath identifying the RPC/action. Any predicates are allowed.
        :arg callback:
            Callback to be called when the RPC/action is invoked.
        :arg priority:
            Specifies the order in which the callbacks (**within RPC/action**) will be
            called.
        :arg no_thread:
            There will be no thread created for handling this subscription meaning no
            event will be processed! Default to True if asyncio_register is True.
        :arg private_data:
            Private context passed to the callback function, opaque to sysrepo.
        :arg asyncio_register:
            Add the created subscription event pipe into asyncio event loop monitored
            read file descriptors. Implies no_thread=True.
        :arg strict:
            Reject the whole data returned by callback if it contains elements without
            schema definition.
        :arg include_implicit_defaults:
            Include implicit defaults into input parameters passed to callbacks.
        :arg extra_info:
            When True, the given callback is called with extra keyword arguments
            containing extra information of the sysrepo session that gave origin to the
            event (see RpcCallbackType for more details)
        """
        if self.is_implicit:
            raise SysrepoUnsupportedError("cannot subscribe with implicit sessions")
        _check_subscription_callback(callback, self.RpcCallbackType)

        sub = Subscription(
            callback,
            private_data,
            asyncio_register=asyncio_register,
            strict=strict,
            include_implicit_defaults=include_implicit_defaults,
            extra_info=extra_info,
        )
        sub_p = ffi.new("sr_subscription_ctx_t **")

        if asyncio_register:
            no_thread = True  # we manage our own event loop
        flags = _subscribe_flags(no_thread=no_thread)

        check_call(
            lib.sr_rpc_subscribe_tree,
            self.cdata,
            str2c(xpath),
            lib.srpy_rpc_tree_cb,
            sub.handle,
            priority,
            flags,
            sub_p,
        )
        sub.init(sub_p[0])

        self.subscriptions.append(sub)

    NotificationCallbackType = Callable[[str, str, Dict, int, Any], None]
    """
    Callback to be called when a notification is received.

    :arg xpath:
        The full xpath to the received notification.
    :arg notification_type:
        Type of the notification event. Can be one of: "realtime", "replay",
        "replay_complete", "stop", "suspended", "resumed".
    :arg notification:
        The notification as a python dictionary. For example, with a YANG notification
        defined like this::

            notification some-notification {
              leaf param1 {
                type uint32;
              }
              leaf param2 {
                type string;
              }
            }

        The notification dict may look like this::

            {'param1': 42, 'param2': 'foobar'}

    :arg timestamp:
        Timestamp of the notification as an unsigned 32-bits integer.
    :arg private_data:
        Private context opaque to sysrepo used when subscribing.
    :arg kwargs (optional):
        If the callback was registered with the argument extra_info=True (see
        Session.subscribe_module_change), then extra keyword arguments are passed when
        calling the callback:
            * netconf_id: the NETCONF session ID set for the event originator
                sysrepo session
            * user: the effective username of the event originator sysrepo session
    """

    def subscribe_notification(
        self,
        module: str,
        xpath: str,
        callback: NotificationCallbackType,
        *,
        start_time: int = 0,
        stop_time: int = 0,
        no_thread: bool = False,
        asyncio_register: bool = False,
        private_data: Any = None,
        extra_info: bool = False,
    ) -> None:
        """
        Subscribe for the delivery of a notification.

        :arg module:
            Name of the module whose notifications to subscribe to.
        :arg xpath:
            XPath identifying the notification.
        :arg callback:
            Callback to be called when the notification is received.
        :arg start_time:
            Optional start time of the subscription. Used for replaying stored
            notifications.
        :arg stop_time:
            Optional stop time ending the notification subscription.
        :arg no_thread:
            There will be no thread created for handling this subscription meaning no
            event will be processed! Default to True if asyncio_register is True.
        :arg asyncio_register:
            Add the created subscription event pipe into asyncio event loop monitored
            read file descriptors. Implies no_thread=True.
        :arg private_data:
            Private context passed to the callback function, opaque to sysrepo.
        :arg extra_info:
            When True, the given callback is called with extra keyword arguments
            containing extra information of the sysrepo session that gave origin to the
            event (see RpcCallbackType for more details)
        """

        if self.is_implicit:
            raise SysrepoUnsupportedError("cannot subscribe with implicit sessions")
        _check_subscription_callback(callback, self.NotificationCallbackType)

        sub = Subscription(
            callback,
            private_data,
            asyncio_register=asyncio_register,
            extra_info=extra_info,
        )

        sub_p = ffi.new("sr_subscription_ctx_t **")

        if asyncio_register:
            no_thread = True  # we manage our own event loop

        flags = _subscribe_flags(no_thread=no_thread)

        c_start_time = ffi.new("struct timespec *")
        c_start_time.tv_sec = start_time
        c_stop_time = ffi.new("struct timespec *")
        c_stop_time.tv_sec = stop_time

        check_call(
            lib.sr_notif_subscribe_tree,
            self.cdata,
            str2c(module),
            str2c(xpath),
            c_start_time,
            c_stop_time,
            lib.srpy_event_notif_tree_cb,
            sub.handle,
            flags,
            sub_p,
        )
        sub.init(sub_p[0])

        self.subscriptions.append(sub)

    # end: subscription

    # begin: changes
    def get_changes(
        self,
        xpath: str,
        include_implicit_defaults: bool = True,
        include_deleted_values: bool = False,
    ) -> Iterator[Change]:
        """
        Return an iterator that will yield all pending changes in the current session.

        :arg xpath:
            Xpath selecting the requested changes. Be careful, you must select all the
            changes, not just subtrees! To get a full change subtree `//.` can be
            appended to the XPath.
        :arg include_implicit_defaults:
            Include implicit default nodes.
        :arg include_deleted_values:
            Include deleted node values into the returned Change objects.

        :returns:
            An iterator that will yield `sysrepo.Change` objects.
        """
        iter_p = ffi.new("sr_change_iter_t **")

        check_call(lib.sr_get_changes_iter, self.cdata, str2c(xpath), iter_p)

        op_p = ffi.new("sr_change_oper_t *")
        node_p = ffi.new("struct lyd_node **")
        prev_val_p = ffi.new("char **")
        prev_list_p = ffi.new("char **")
        prev_dflt_p = ffi.new("int *")
        try:
            ret = check_call(
                lib.sr_get_change_tree_next,
                self.cdata,
                iter_p[0],
                op_p,
                node_p,
                prev_val_p,
                prev_list_p,
                prev_dflt_p,
                valid_codes=(lib.SR_ERR_OK, lib.SR_ERR_NOT_FOUND),
            )
            while ret == lib.SR_ERR_OK:
                try:
                    with self.get_ly_ctx() as ctx:
                        yield Change.parse(
                            operation=op_p[0],
                            node=libyang.DNode.new(ctx, node_p[0]),
                            prev_val=c2str(prev_val_p[0]),
                            prev_list=c2str(prev_list_p[0]),
                            prev_dflt=bool(prev_dflt_p[0]),
                            include_implicit_defaults=include_implicit_defaults,
                            include_deleted_values=include_deleted_values,
                        )
                except Change.Skip:
                    pass
                ret = check_call(
                    lib.sr_get_change_tree_next,
                    self.cdata,
                    iter_p[0],
                    op_p,
                    node_p,
                    prev_val_p,
                    prev_list_p,
                    prev_dflt_p,
                    valid_codes=(lib.SR_ERR_OK, lib.SR_ERR_NOT_FOUND),
                )
        finally:
            lib.sr_free_change_iter(iter_p[0])

    # end: changes

    # begin: get
    def get_item(self, xpath: str, timeout_ms: int = 0) -> Value:
        """
        Retrieve a single data element selected by the provided path.

        :arg xpath:
            Path of the data element to be retrieved.
        :arg timeout_ms:
            Operational callback timeout in milliseconds. If 0, default is used.

        :returns:
            A sysrepo.Value object.
        :raises SysrepoInvalArgError:
            If multiple nodes match the path.
        :raises SysrepoNotFoundError:
            If no nodes match the path.
        """
        val_p = ffi.new("sr_val_t **")
        check_call(lib.sr_get_item, self.cdata, str2c(xpath), timeout_ms, val_p)
        try:
            return Value.parse(val_p[0])
        finally:
            lib.sr_free_val(val_p[0])

    def get_items(
        self,
        xpath: str,
        timeout_ms: int = 0,
        no_state: bool = False,
        no_config: bool = False,
        no_subs: bool = False,
        no_stored: bool = False,
    ) -> Iterator[Value]:
        """
        Retrieve an array of data elements selected by the provided XPath.

        All data elements are transferred within one message from the datastore, which
        is more efficient than multiple get_item calls.

        :arg xpath:
            XPath of the data elements to be retrieved.
        :arg timeout_ms:
            Operational callback timeout in milliseconds. If 0, default is used.
        :arg no_state:
            Return only configuration data.
        :arg no_config:
            Return only state data. If there are some state subtrees with configuration
            these are also returned (with keys if lists).
        :arg no_subs:
            Return only stored operational data (push), do not call subscriber
            callbacks (pull).
        :arg no_stored:
            Do not merge with stored operational data (push).

        :returns:
            An iterator that yields sysrepo.Value objects.
        """
        flags = _get_oper_flags(
            no_state=no_state, no_config=no_config, no_subs=no_subs, no_stored=no_stored
        )
        val_p = ffi.new("sr_val_t **")
        count_p = ffi.new("size_t *")
        check_call(
            lib.sr_get_items,
            self.cdata,
            str2c(xpath),
            timeout_ms,
            flags,
            val_p,
            count_p,
        )
        try:
            for i in range(count_p[0]):
                yield Value.parse(val_p[0] + i)
        finally:
            lib.sr_free_values(val_p[0], count_p[0])

    @contextmanager
    def get_data_ly(
        self,
        xpath: str,
        max_depth: int = 0,
        timeout_ms: int = 0,
        no_state: bool = False,
        no_config: bool = False,
        no_subs: bool = False,
        no_stored: bool = False,
    ) -> libyang.DNode:
        """
        Retrieve a tree whose root nodes match the provided XPath.

        Top-level trees are always returned so if an inner node is selected, all of its
        descendants and its direct parents (lists also with keys) are returned.

        If the subtree selection process results in too many node overlaps, the cost of
        the operation may be unnecessarily big. As an example, a common XPath expression
        `//.` is normally used to select all nodes in a data tree, but for this
        operation it would result in an excessive duplication of data nodes. Since all
        the descendants of each matched node are returned implicitly, `//` in the XPath
        should never be used (i.e. `/*` is the correct XPath for all the nodes).

        :arg xpath:
            Path selecting the root nodes of subtrees to be retrieved.
        :arg max_depth:
            Maximum depth of the selected subtrees. 0 is unlimited, 1 will not return
            any descendant nodes. If a list should be returned, its keys are always
            returned as well.
        :arg timeout_ms:
            Operational callback timeout in milliseconds. If 0, default is used.
        :arg no_state:
            Return only configuration data.
        :arg no_config:
            Return only state data. If there are some state subtrees with configuration
            these are also returned (with keys if lists).
        :arg no_subs:
            Return only stored operational data (push), do not call subscriber callbacks
            (pull).
        :arg no_stored:
            Do not merge with stored operational data (push).

        :returns:
            A libyang.DNode object with all the requested data, allocated dynamically.
            It must be freed manually by the caller with the libyang.DNode.free()
            method.
        :raises SysrepoNotFoundError:
            If no nodes match the path.
        """
        flags = _get_oper_flags(
            no_state=no_state, no_config=no_config, no_subs=no_subs, no_stored=no_stored
        )
        if flags and lib.sr_session_get_ds(self.cdata) != lib.SR_DS_OPERATIONAL:
            raise ValueError(
                '"no_*" arguments are only valid for the "operational" datastore'
            )
        sr_data_p = ffi.new("sr_data_t **")
        check_call(
            lib.sr_get_data,
            self.cdata,
            str2c(xpath),
            max_depth,
            timeout_ms,
            flags,
            sr_data_p,
        )
        if not sr_data_p[0]:
            raise SysrepoNotFoundError(xpath)

        ctx = self.acquire_context()
        try:
            dnode = libyang.DNode.new(ctx, sr_data_p[0].tree).root()

            # customize the free method to use the sysrepo free
            def sysrepo_free(dnode_src):
                lib.sr_release_data(sr_data_p[0])
                self.release_context()

            dnode.free_func = sysrepo_free

            yield dnode
        finally:
            dnode.free()

    def new_dnode(self, cdata) -> libyang.DNode:
        """
        Create a new DNode which release the context.
        """
        ctx = self.acquire_context()
        new_dnode = libyang.DNode.new(ctx, cdata)

        def free_func(dnode):
            self.release_context()
            dnode.free_internal()

        new_dnode.free_func = free_func

        return new_dnode

    def get_data(
        self,
        xpath: str,
        max_depth: int = 0,
        timeout_ms: int = 0,
        no_state: bool = False,
        no_config: bool = False,
        no_subs: bool = False,
        no_stored: bool = False,
        strip_prefixes: bool = True,
        include_implicit_defaults: bool = False,
        trim_default_values: bool = False,
        keep_empty_containers: bool = False,
    ) -> Dict:
        """
        Same as `SysrepoSession.get_data_ly` but returns a python dictionary.

        :arg strip_prefixes:
            If True, remove YANG module prefixes from dictionary keys.

        :returns:
            A python dictionary generated from the returned struct lyd_node.
        """
        with self.get_data_ly(
            xpath,
            max_depth=max_depth,
            timeout_ms=timeout_ms,
            no_state=no_state,
            no_config=no_config,
            no_subs=no_subs,
            no_stored=no_stored,
        ) as data:
            return data.print_dict(
                with_siblings=True,
                absolute=True,
                strip_prefixes=strip_prefixes,
                include_implicit_defaults=include_implicit_defaults,
                trim_default_values=trim_default_values,
                keep_empty_containers=keep_empty_containers,
            )

    # end: get

    # begin: edit
    def set_item(self, xpath: str, value: Any = None) -> None:
        """
        Prepare to set (create) the value of a leaf, leaf-list, list, or presence
        container. These changes are applied only after calling apply_changes().

        :arg xpath:
            Path identifier of the data element to be set.
        :arg value:
            Value to be set. It will be converted to string.
        """
        if value is not None and not isinstance(value, str):
            if isinstance(value, bool):
                value = str(value).lower()
            else:
                value = str(value)
        check_call(
            lib.sr_set_item_str, self.cdata, str2c(xpath), str2c(value), ffi.NULL, 0
        )

    def discard_items(self, xpath: str) -> None:
        """
        Prepare to discard nodes matching the specified xpath (or all if not
        set) previously set by the session connection. Usable only for the
        operational datastore . These changes are applied only after calling
        apply_changes().

        :arg xpath:
            Path identifier of the data element to be deleted.
        """
        check_call(lib.sr_discard_items, self.cdata, str2c(xpath))

    def delete_item(self, xpath: str) -> None:
        """
        Prepare to delete the nodes matching the specified xpath. These changes
        are applied only after calling apply_changes().

        :arg xpath:
            Path identifier of the data element to be deleted.

        :raises SysrepoNotFoundError:
            If no nodes match the path.
        """
        check_call(lib.sr_delete_item, self.cdata, str2c(xpath), 0)

    def delete_oper_item(self, xpath: str, value: Any = None) -> None:
        """
        Prepare to delete the nodes matching the specified xpath in the operational
        datastore. These changes are applied only after calling
        apply_changes().

        :arg xpath:
            Path identifier of the data element to be deleted.
        :arg value:
            String representation of the value deleted.

        :raises SysrepoNotFoundError:
            If no nodes match the path.
        """
        if value is not None and not isinstance(value, str):
            if isinstance(value, bool):
                value = str(value).lower()
            else:
                value = str(value)
        check_call(
            lib.sr_oper_delete_item_str, self.cdata, str2c(xpath), str2c(value), 0
        )

    def edit_batch_ly(
        self, edit: libyang.DNode, default_operation: str = "merge"
    ) -> None:
        """
        Provide a prepared edit data tree to be applied. These changes are
        applied only after calling apply_changes().

        :arg edit:
            Data tree holding the configuration. Similar semantics to
            https://tools.ietf.org/html/rfc6241#section-7.2. The given data tree is
            *NOT* spent and must be freed by the caller.
        :arg default_operation:
            Default operation for nodes without operation on themselves or any parent.
            Possible values are `merge`, `replace`, or `none`. See RFC 6241
            https://tools.ietf.org/html/rfc6241#page-39.
        """
        # libyang and sysrepo bindings are different, casting is required
        dnode = ffi.cast("struct lyd_node *", edit.cdata)
        check_call(lib.sr_edit_batch, self.cdata, dnode, str2c(default_operation))

    def edit_batch(
        self,
        edit: Dict,
        module_name: str,
        strict: bool = False,
        default_operation: str = "merge",
    ) -> None:
        """
        Same as `SysrepoSession.edit_batch_ly` but with a python dictionary.

        :arg edit:
            Python dictionary holding the configuration.
        :arg module_name:
            The YANG module name matching the data in the dictionary. It is used to
            convert the dictionary to a `libyang.DNode` object.
        :arg strict:
            If True, reject config if it contains elements without any schema
            definition.
        """
        with self.get_ly_ctx() as ctx:
            module = ctx.get_module(module_name)

        dnode = module.parse_data_dict(edit, strict=strict, validate=False)
        if not dnode:
            raise ValueError("provided config dict is empty")
        try:
            self.edit_batch_ly(dnode, default_operation)
        finally:
            dnode.free()

    def replace_config_ly(
        self,
        config: Optional[libyang.DNode],
        module_name: Optional[str],
        timeout_ms: int = 0,
    ) -> None:
        """
        Replace a datastore with the contents of a data tree.

        :arg config:
            Source data to replace the datastore. Is ALWAYS spent and cannot be further
            used by the application! Can be None to completely reset the configuration.
        :arg module_name:
            The module for which to replace the configuration.
        :arg timeout_ms:
            Configuration callback timeout in milliseconds. If 0, default is used.

        :raises SysrepoError:
            If the operation failed.
        """
        if isinstance(config, libyang.DNode):
            # libyang and sysrepo bindings are different, casting is required
            dnode = ffi.cast("struct lyd_node *", config.cdata)
        elif config is None:
            dnode = ffi.NULL
        else:
            raise TypeError("config must be either a libyang.DNode object or None")
        check_call(
            lib.sr_replace_config,
            self.cdata,
            str2c(module_name),
            dnode,
            timeout_ms,
        )

    def replace_config(
        self,
        config: Optional[Dict],
        module_name: str,
        strict: bool = False,
        timeout_ms: int = 0,
    ) -> None:
        """
        Same as replace_config() but with a python dictionary.

        :arg config:
            The configuration dict.
        :arg strict:
            If True, reject config if it contains elements without any schema
            definition.
        """
        with self.get_ly_ctx() as ctx:
            module = ctx.get_module(module_name)

        dnode = module.parse_data_dict(config, strict=strict, validate=False)
        self.replace_config_ly(dnode, module_name, timeout_ms=timeout_ms)

    def copy_config(
        self, src_datastore: str, module_name: str = None, timeout_ms: int = 0
    ) -> None:
        """
        Replaces a conventional datastore with the contents of another conventional
        datastore. If the module is specified, limits the operation only to the
        specified module. If it is not specified, the operation is performed on all
        modules.

        Note that copying from candidate to running or vice versa causes the candidate
        datastore to revert to original behavior of mirroring running datastore
        (datastores).

        Required WRITE access.

        :arg src_datastore:
            Source datastore to copy configuration from. Can be one of `running`,
            `startup`, `operational` or `candidate`. Must be a different datastore
            than the session's.
        :arg module_name:
            Optional module name that limits the copy operation only to this module.
        :arg timeout_ms:
            Configuration callback timeout in milliseconds. If 0, default is used.
        """
        if self.is_implicit:
            raise SysrepoUnsupportedError("cannot copy config from implicit sessions")
        ds = datastore_value(src_datastore)
        check_call(lib.sr_copy_config, self.cdata, str2c(module_name), ds, timeout_ms)

    def validate(self, module_name: str = None) -> None:
        """
        Perform the validation a datastore and any changes made in the current
        session, but do not apply nor discard them.

        Provides only YANG validation, apply-changes **subscribers will not be
        notified** in this case.

        :arg module_name:
            If specified, limits the validate operation only to this module and its
            dependencies.

        :raises SysrepoError:
            If validation failed.
        """
        if self.is_implicit:
            raise SysrepoUnsupportedError("cannot validate with implicit sessions")
        check_call(lib.sr_validate, self.cdata, str2c(module_name), 0)

    def apply_changes(self, timeout_ms: int = 0) -> None:
        """
        Apply changes made in the current session. In case the changes could not be
        applied successfully for any reason, they remain intact in the session until
        discard_changes() is called.

        :arg timeout_ms:
            Configuration callback timeout in milliseconds. If 0, default is used.

        :raises SysrepoError:
            If changes could not be applied.
        """
        if self.is_implicit:
            raise SysrepoUnsupportedError("cannot apply_changes with implicit sessions")
        check_call(lib.sr_apply_changes, self.cdata, timeout_ms)

    def discard_changes(self) -> None:
        """
        Discard prepared changes made in the current session.
        """
        if self.is_implicit:
            raise SysrepoUnsupportedError(
                "cannot discard_changes with implicit sessions"
            )
        check_call(lib.sr_discard_changes, self.cdata)

    # end: edit

    def rpc_send_ly(
        self, rpc_input: libyang.DNode, timeout_ms: int = 0
    ) -> Optional[libyang.DNode]:
        """
        Send an RPC/action and wait for the result.

        RPC/action must be valid in (is validated against) the operational datastore
        context.

        :arg rpc_input:
            The RPC/action input tree. It is *NOT* spent and must be freed by the
            caller.
        :arg timeout_ms:
            RPC/action callback timeout in milliseconds. If 0, default is used.

        :returns:
            The RPC/action output tree. Allocated dynamically and must be freed by the
            caller. If the RPC/action did not return any output, None is returned.
        :raises SysrepoError:
            If the RPC/action callback failed.
        """
        if not isinstance(rpc_input, libyang.DNode):
            raise TypeError("rpc_input must be a libyang.DNode")
        # libyang and sysrepo bindings are different, casting is required
        in_dnode = ffi.cast("struct lyd_node *", rpc_input.cdata)
        sr_data_p = ffi.new("sr_data_t **")
        check_call(lib.sr_rpc_send_tree, self.cdata, in_dnode, timeout_ms, sr_data_p)
        if not sr_data_p[0]:
            return None

        ctx = self.acquire_context()
        dnode = libyang.DNode.new(ctx, sr_data_p[0].tree)

        # customize the free method to use the sysrepo free
        def sysrepo_free(dnode_src):
            lib.sr_release_data(sr_data_p[0])
            self.release_context()

        dnode.free_func = sysrepo_free

        return dnode

    def rpc_send(
        self,
        xpath: str,
        input_dict: Dict,
        timeout_ms: int = 0,
        strict: bool = False,
        strip_prefixes: bool = True,
        include_implicit_defaults: bool = False,
        trim_default_values: bool = False,
        keep_empty_containers: bool = False,
    ) -> Optional[Dict]:
        """
        Same as rpc_send_ly() but takes a python dictionary and a YANG module name as
        input arguments.

        :arg rpc_input:
            Input data tree. It is converted to a libyang struct lyd_node according to
            module_name.
        :arg module_name:
            The name of the YANG module used to convert the rpc input dict to a
            libyang.DNode object.
        :arg strict:
            If True, reject rpc_input if it contains elements without any schema
            definition.
        :arg strip_prefixes:
            If True, remove YANG module prefixes from dictionary keys.
        :arg include_implicit_defaults:
            Include leaves with implicit default values in the returned dict.
        :arg trim_default_values:
            Exclude leaves when their value equals the default.
        :arg keep_empty_containers:
            Include empty (non-presence) containers.

        :returns:
            A python dictionary with the RPC/action output tree or None if the RPC/action
            did not return any output.
        """
        rpc = {}
        libyang.xpath_set(rpc, xpath, input_dict)
        module_name, _, _ = next(libyang.xpath_split(xpath))
        with self.get_ly_ctx() as ctx:
            module = ctx.get_module(module_name)
        in_dnode = module.parse_data_dict(rpc, rpc=True, strict=strict, validate=False)

        try:
            out_dnode = self.rpc_send_ly(in_dnode, timeout_ms=timeout_ms)
        finally:
            in_dnode.free()

        if out_dnode is None:
            return None

        try:
            out_dict = out_dnode.print_dict(
                strip_prefixes=strip_prefixes,
                absolute=False,
                with_siblings=False,
                include_implicit_defaults=include_implicit_defaults,
                trim_default_values=trim_default_values,
                keep_empty_containers=keep_empty_containers,
            )
            # strip first item with only RPC/action name
            return next(iter(out_dict.values()))
        finally:
            out_dnode.free()

    def notification_send_ly(
        self, notification: libyang.DNode, timeout_ms: int = 0, wait: bool = False
    ) -> None:
        """
        Send a notification

        :arg notification:
            The notification tree. It is *NOT* spent and must be freed by the
            caller.
        :arg timeout_ms:
             Configuration callback timeout in milliseconds. If 0, default is used.
        :arg wait:
            If True, wait until all callbacks on all events are finished (even "done" or
            "abort"). If not set, these events may not yet be processed after the
            function returns. Note that all "change" events are always waited for.

        :raises SysrepoError:
            If the notification callback failed.
        """
        if not isinstance(notification, libyang.DNode):
            raise TypeError("notification must be a libyang.DNode")
        # libyang and sysrepo bindings are different, casting is required
        in_dnode = ffi.cast("struct lyd_node *", notification.cdata)
        check_call(lib.sr_notif_send_tree, self.cdata, in_dnode, timeout_ms, wait)

    def notification_send(
        self, xpath: str, notification: Dict, strict: bool = False
    ) -> None:
        """
        Same as notification_send_ly() but takes a python dictionary as a notification.

        :arg xpath:
            The xpath corresponding to the notification
        :arg notification:
            The notification to send represented as a dictionary
        :arg strict:
            If True, reject notification if it contains elements without any schema
            definition.

        :raises SysrepoError:
            If the notification callback failed.
        """

        full_notification = {}
        libyang.xpath_set(full_notification, xpath, notification)
        module_name, _, _ = next(libyang.xpath_split(xpath))
        with self.get_ly_ctx() as ctx:
            module = ctx.get_module(module_name)
        dnode = module.parse_data_dict(
            full_notification, notification=True, strict=strict, validate=False
        )
        try:
            self.notification_send_ly(dnode)
        finally:
            dnode.free()


# -------------------------------------------------------------------------------------
DATASTORE_VALUES = {
    "running": lib.SR_DS_RUNNING,
    "operational": lib.SR_DS_OPERATIONAL,
    "startup": lib.SR_DS_STARTUP,
    "candidate": lib.SR_DS_CANDIDATE,
}


def datastore_value(name: str) -> int:
    if name not in DATASTORE_VALUES:
        raise ValueError("unknown datastore name: %r" % name)
    return DATASTORE_VALUES[name]


def datastore_name(value: int) -> str:
    for name, val in DATASTORE_VALUES.items():
        if val == value:
            return name
    raise ValueError("unknown datastore value: %r" % value)


# -------------------------------------------------------------------------------------
def _get_oper_flags(no_state=False, no_config=False, no_subs=False, no_stored=False):
    flags = 0
    if no_state:
        flags |= lib.SR_OPER_NO_STATE
    if no_config:
        flags |= lib.SR_OPER_NO_CONFIG
    if no_subs:
        flags |= lib.SR_OPER_NO_SUBS
    if no_stored:
        flags |= lib.SR_OPER_NO_STORED
    return flags


# -------------------------------------------------------------------------------------
def _subscribe_flags(
    no_thread=False,
    passive=False,
    done_only=False,
    enabled=False,
    oper_merge=False,
    filter_origin=False,
    update=False,
):
    flags = 0
    if no_thread:
        flags |= lib.SR_SUBSCR_NO_THREAD
    if passive:
        flags |= lib.SR_SUBSCR_PASSIVE
    if done_only:
        flags |= lib.SR_SUBSCR_DONE_ONLY
    if enabled:
        flags |= lib.SR_SUBSCR_ENABLED
    if update:
        flags |= lib.SR_SUBSCR_UPDATE
    if oper_merge:
        flags |= lib.SR_SUBSCR_OPER_MERGE
    if filter_origin:
        flags |= lib.SR_SUBSCR_FILTER_ORIG
    return flags


# -------------------------------------------------------------------------------------
def _check_subscription_callback(callback, expected_type):
    if not inspect.isroutine(callback):
        raise TypeError("callback must be a function")
    *arg_types, return_type = expected_type.__args__
    sig = inspect.signature(callback)
    callback_positional_args = tuple(
        p
        for p in sig.parameters.values()
        if p.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD
    )
    if len(arg_types) != len(callback_positional_args):
        raise ValueError(
            "callback %s does not have required arguments: (%s) -> %s"
            % (callback, arg_types, return_type)
        )



#!/usr/bin/env python3
# Copyright (c) 2020 6WIND S.A.
# SPDX-License-Identifier: BSD-3-Clause

import datetime
import os
import re
import subprocess

import setuptools
import setuptools.command.sdist


# ------------------------------------------------------------------------------
def git_describe_to_pep440(version):
    """
    ``git describe`` produces versions in the form: `v0.9.8-20-gf0f45ca` where
    20 is the number of commit since last release, and gf0f45ca is the short
    commit id preceded by 'g' we parse this a transform into a pep440 release
    version 0.9.9.dev20 (increment last digit and add dev before 20)
    """
    match = re.search(
        r"""
        v(?P<major>\d+)\.
        (?P<minor>\d+)\.
        (?P<patch>\d+)
        ((\.post|-)(?P<post>\d+)(?!-g))?
        ([\+~](?P<local_segment>.*?))?
        (-(?P<dev>\d+))?(-g(?P<commit>.+))?
        """,
        version,
        flags=re.VERBOSE,
    )
    if not match:
        raise ValueError("unknown tag format")
    dic = {
        "major": int(match.group("major")),
        "minor": int(match.group("minor")),
        "patch": int(match.group("patch")),
    }
    fmt = "{major}.{minor}.{patch}"
    if match.group("dev"):
        dic["patch"] += 1
        dic["dev"] = int(match.group("dev"))
        fmt += ".dev{dev}"
    elif match.group("post"):
        dic["post"] = int(match.group("post"))
        fmt += ".post{post}"
    if match.group("local_segment"):
        dic["local_segment"] = match.group("local_segment")
        fmt += "+{local_segment}"
    return fmt.format(**dic)


# ------------------------------------------------------------------------------
def get_version_from_archive_id(git_archive_id="$Format:%ct %d$"):
    """
    Extract the tag if a source is from git archive.

    When source is exported via `git archive`, the git_archive_id init value is
    modified and placeholders are expanded to the "archived" revision:

        %ct: committer date, UNIX timestamp
        %d: ref names, like the --decorate option of git-log

    See man gitattributes(5) and git-log(1) (PRETTY FORMATS) for more details.
    """
    # mangle the magic string to make sure it is not replaced by git archive
    if git_archive_id.startswith("$For" "mat:"):  # pylint: disable=implicit-str-concat
        raise ValueError("source was not modified by git archive")

    # source was modified by git archive, try to parse the version from
    # the value of git_archive_id
    match = re.search(r"tag:\s*([^,)]+)", git_archive_id)
    if match:
        # archived revision is tagged, use the tag
        return git_describe_to_pep440(match.group(1))

    # archived revision is not tagged, use the commit date
    tstamp = git_archive_id.strip().split()[0]
    d = datetime.datetime.utcfromtimestamp(int(tstamp))
    return d.strftime("1.%Y.%m.%d")


# ------------------------------------------------------------------------------
def read_file(fpath, encoding="utf-8"):
    with open(fpath, "r", encoding=encoding) as f:
        return f.read().strip()


# ------------------------------------------------------------------------------
def get_version():
    try:
        return read_file("sysrepo/VERSION")
    except IOError:
        pass

    if "SYSREPO_PYTHON_FORCE_VERSION" in os.environ:
        return os.environ["SYSREPO_PYTHON_FORCE_VERSION"]

    try:
        return get_version_from_archive_id()
    except ValueError:
        pass

    try:
        out = subprocess.check_output(
            ["git", "describe", "--tags", "--always"], stderr=subprocess.DEVNULL
        )
        return git_describe_to_pep440(out.decode("utf-8").strip())
    except Exception:
        pass

    return "1.999999.999999"


# ------------------------------------------------------------------------------
class SDistCommand(setuptools.command.sdist.sdist):
    def write_lines(self, file, lines):
        with open(file, "w", encoding="utf-8") as f:
            for line in lines:
                f.write(line + "\n")

    def make_release_tree(self, base_dir, files):
        super().make_release_tree(base_dir, files)
        version_file = os.path.join(base_dir, "sysrepo/VERSION")
        self.execute(
            self.write_lines,
            (version_file, [self.distribution.metadata.version]),
            "Writing %s" % version_file,
        )


# ------------------------------------------------------------------------------
setuptools.setup(
    name="sysrepo",
    version=get_version(),
    description="Sysrepo CFFI bindings",
    long_description=read_file("README.rst"),
    license="BSD 3 clause",
    author="Robin Jarry",
    author_email="robin.jarry@6wind.com",
    url="https://www.6wind.com/",
    keywords=["sysrepo", "libyang", "cffi"],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: Unix",
        "Programming Language :: Python :: 3",
        "Topic :: Software Development :: Libraries",
    ],
    packages=["sysrepo"],
    python_requires=">=3.6",
    setup_requires=[
        "setuptools",
        'cffi; platform_python_implementation != "PyPy"',
    ],
    install_requires=[
        "libyang>=2.1",
        'cffi; platform_python_implementation != "PyPy"',
    ],
    cffi_modules=["cffi/build.py:BUILDER"],
    include_package_data=True,
    zip_safe=False,
    cmdclass={"sdist": SDistCommand},
)



# Copyright (c) 2020 6WIND S.A.
# SPDX-License-Identifier: BSD-3-Clause

import getpass
import logging
import os
import unittest

import sysrepo
from sysrepo.session import SysrepoSession


YANG_FILE = os.path.join(
    os.path.dirname(os.path.dirname(__file__)), "examples/sysrepo-example.yang"
)
sysrepo.configure_logging(stderr_level=logging.ERROR)


# ------------------------------------------------------------------------------
class ModuleChangeSubscriptionTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.conn = sysrepo.SysrepoConnection()
        cls.conn.install_module(YANG_FILE, enabled_features=["turbo"])

    @classmethod
    def tearDownClass(cls):
        # we have to disconnect first to release all resources
        cls.conn.disconnect()
        with sysrepo.SysrepoConnection() as c:
            c.remove_module("sysrepo-example")

    def setUp(self):
        with self.conn.start_session("running") as sess:
            sess.delete_item("/sysrepo-example:conf")
            sess.apply_changes()
        self.sess = self.conn.start_session()

    def tearDown(self):
        self.sess.stop()

    def test_module_change_sub(self):
        priv = object()
        current_config = {}
        expected_changes = []

        def module_change_cb(event, req_id, changes, private_data):
            self.assertIn(event, ("change", "done", "abort", "update"))
            self.assertIs(private_data, priv)
            for c in changes:
                if c.xpath == "/sysrepo-example:conf/system/hostname":
                    if event == "change" and c.value == "INVALID":
                        raise sysrepo.SysrepoValidationFailedError("invalid hostname")
            if event in ("change", "done"):
                self.assertEqual(changes, expected_changes)
            if event == "done":
                sysrepo.update_config_cache(current_config, changes)

        self.sess.subscribe_module_change(
            "sysrepo-example",
            "/sysrepo-example:conf",
            module_change_cb,
            private_data=priv,
            update=True,
        )

        with self.conn.start_session("running") as ch_sess:
            # 1.
            sent_config = {"conf": {"system": {"hostname": "bar"}}}
            expected_changes = [
                sysrepo.ChangeCreated("/sysrepo-example:conf/system/hostname", "bar")
            ]
            ch_sess.replace_config(sent_config, "sysrepo-example", strict=True)
            self.assertEqual(current_config, sent_config)
            # 2.
            sent_config = {"conf": {"system": {"hostname": "INVALID"}}}
            expected_changes = []
            with self.assertRaises(sysrepo.SysrepoValidationFailedError):
                ch_sess.replace_config(sent_config, "sysrepo-example", strict=True)
            # 3.
            sent_config = {
                "conf": {
                    "system": {"hostname": "bar"},
                    "network": {"interface": [{"name": "eth0", "up": True}]},
                }
            }
            expected_changes = [
                sysrepo.ChangeCreated(
                    "/sysrepo-example:conf/network/interface[name='eth0']",
                    {"name": "eth0", "up": True},
                    after="",
                ),
                sysrepo.ChangeCreated(
                    "/sysrepo-example:conf/network/interface[name='eth0']/name", "eth0"
                ),
                sysrepo.ChangeCreated(
                    "/sysrepo-example:conf/network/interface[name='eth0']/up", True
                ),
            ]
            ch_sess.replace_config(sent_config, "sysrepo-example", strict=True)
            self.assertEqual(current_config, sent_config)
            # 4.
            sent_config = {
                "conf": {
                    "system": {"hostname": "bar"},
                    "network": {"interface": [{"name": "eth2", "up": False}]},
                }
            }
            expected_changes = [
                sysrepo.ChangeDeleted(
                    "/sysrepo-example:conf/network/interface[name='eth0']",
                    None,
                ),
                sysrepo.ChangeDeleted(
                    "/sysrepo-example:conf/network/interface[name='eth0']/name",
                    None,
                ),
                sysrepo.ChangeDeleted(
                    "/sysrepo-example:conf/network/interface[name='eth0']/up",
                    None,
                ),
                sysrepo.ChangeCreated(
                    "/sysrepo-example:conf/network/interface[name='eth2']",
                    {"name": "eth2", "up": False},
                    after="",
                ),
                sysrepo.ChangeCreated(
                    "/sysrepo-example:conf/network/interface[name='eth2']/name", "eth2"
                ),
                sysrepo.ChangeCreated(
                    "/sysrepo-example:conf/network/interface[name='eth2']/up", False
                ),
            ]
            ch_sess.replace_config(sent_config, "sysrepo-example", strict=True)
            self.assertEqual(current_config, sent_config)
            # 5.
            sent_config = {
                "conf": {
                    "system": {"hostname": "bar"},
                    "network": {"interface": [{"name": "eth2", "up": True}]},
                }
            }
            expected_changes = [
                sysrepo.ChangeModified(
                    "/sysrepo-example:conf/network/interface[name='eth2']/up",
                    True,
                    "false",
                    False,
                ),
            ]
            ch_sess.replace_config(sent_config, "sysrepo-example", strict=True)
            self.assertEqual(current_config, sent_config)
            # 6.
            sent_config = {
                "conf": {
                    "system": {"hostname": "bar"},
                    "network": {
                        "interface": [
                            {"name": "eth2", "up": True},
                            {"name": "eth0", "up": False},
                        ]
                    },
                }
            }
            expected_changes = [
                sysrepo.ChangeCreated(
                    "/sysrepo-example:conf/network/interface[name='eth0']",
                    {"name": "eth0", "up": False},
                    after="[name='eth2']",
                ),
                sysrepo.ChangeCreated(
                    "/sysrepo-example:conf/network/interface[name='eth0']/name", "eth0"
                ),
                sysrepo.ChangeCreated(
                    "/sysrepo-example:conf/network/interface[name='eth0']/up", False
                ),
            ]
            ch_sess.replace_config(sent_config, "sysrepo-example", strict=True)
            self.assertEqual(current_config, sent_config)
            # 7.
            sent_config = {
                "conf": {
                    "system": {"hostname": "bar"},
                    "network": {
                        "interface": [
                            {"name": "eth0", "up": False},
                            {"name": "eth2", "up": True},
                        ]
                    },
                }
            }
            expected_changes = [
                sysrepo.ChangeMoved(
                    "/sysrepo-example:conf/network/interface[name='eth0']", after=""
                ),
            ]
            ch_sess.replace_config(sent_config, "sysrepo-example", strict=True)
            self.assertEqual(current_config, sent_config)

    def test_module_change_sub_with_deleted_values(self):
        priv = object()
        current_config = {}
        expected_changes = []

        def module_change_cb(event, req_id, changes, private_data):
            self.assertIn(event, ("change", "done", "abort"))
            self.assertIs(private_data, priv)
            if event in ("change", "done"):
                self.assertEqual(changes, expected_changes)
            if event == "done":
                sysrepo.update_config_cache(current_config, changes)

        self.sess.subscribe_module_change(
            "sysrepo-example",
            "/sysrepo-example:conf",
            module_change_cb,
            private_data=priv,
            include_deleted_values=True,
        )

        with self.conn.start_session("running") as ch_sess:
            # 1.
            sent_config = {
                "conf": {
                    "system": {"hostname": "bar"},
                    "network": {"interface": [{"name": "eth0", "up": True}]},
                }
            }
            expected_changes = [
                sysrepo.ChangeCreated("/sysrepo-example:conf/system/hostname", "bar"),
                sysrepo.ChangeCreated(
                    "/sysrepo-example:conf/network/interface[name='eth0']",
                    {"name": "eth0", "up": True},
                    after="",
                ),
                sysrepo.ChangeCreated(
                    "/sysrepo-example:conf/network/interface[name='eth0']/name", "eth0"
                ),
                sysrepo.ChangeCreated(
                    "/sysrepo-example:conf/network/interface[name='eth0']/up", True
                ),
            ]
            ch_sess.replace_config(sent_config, "sysrepo-example", strict=True)
            self.assertEqual(current_config, sent_config)
            # 2.
            sent_config = {
                "conf": {
                    "system": {"hostname": "bar"},
                    "network": {"interface": [{"name": "eth2", "up": False}]},
                }
            }
            expected_changes = [
                sysrepo.ChangeDeleted(
                    "/sysrepo-example:conf/network/interface[name='eth0']",
                    {"name": "eth0", "up": True},
                ),
                sysrepo.ChangeDeleted(
                    "/sysrepo-example:conf/network/interface[name='eth0']/name",
                    "eth0",
                ),
                sysrepo.ChangeDeleted(
                    "/sysrepo-example:conf/network/interface[name='eth0']/up",
                    True,
                ),
                sysrepo.ChangeCreated(
                    "/sysrepo-example:conf/network/interface[name='eth2']",
                    {"name": "eth2", "up": False},
                    after="",
                ),
                sysrepo.ChangeCreated(
                    "/sysrepo-example:conf/network/interface[name='eth2']/name", "eth2"
                ),
                sysrepo.ChangeCreated(
                    "/sysrepo-example:conf/network/interface[name='eth2']/up", False
                ),
            ]
            ch_sess.replace_config(sent_config, "sysrepo-example", strict=True)
            self.assertEqual(current_config, sent_config)

    def test_module_change_sub_with_extra_info(self):
        priv = object()
        calls = []

        def module_change_cb(event, req_id, changes, private_data, **kwargs):
            self.assertIn(event, ("change", "done", "abort"))
            self.assertIsInstance(req_id, int)
            self.assertIsInstance(changes, list)
            self.assertIs(private_data, priv)
            self.assertIn("user", kwargs)
            self.assertEqual(getpass.getuser(), kwargs["user"])
            self.assertIn("netconf_id", kwargs)
            self.assertEqual(12, kwargs["netconf_id"])
            calls.append((event, req_id, changes, private_data, kwargs))

        self.sess.subscribe_module_change(
            "sysrepo-example",
            "/sysrepo-example:conf",
            module_change_cb,
            private_data=priv,
            extra_info=True,
        )

        with self.conn.start_session("running") as ch_sess:
            ch_sess.set_extra_info("netopeer2", 12, getpass.getuser())

            sent_config = {"conf": {"system": {"hostname": "bar"}}}
            ch_sess.replace_config(sent_config, "sysrepo-example", strict=True)
            # Successful change callbacks are called twice:
            #   * once with event "change"
            #   * once with event "done"
            self.assertEqual(2, len(calls))

    def test_module_change_sub_unsafe(self):
        priv = object()
        current_config = {}
        expected_changes = []

        def module_change_cb(session, event, req_id, private_data):
            self.assertIsInstance(session, SysrepoSession)
            self.assertIn(event, ("change", "done", "abort", "update"))
            self.assertIsInstance(req_id, int)
            self.assertIs(private_data, priv)
            changes = list(session.get_changes("/sysrepo-example:conf//."))
            for c in changes:
                if c.xpath == "/sysrepo-example:conf/system/hostname":
                    if event == "change" and c.value == "INVALID":
                        raise sysrepo.SysrepoValidationFailedError("invalid hostname")
            if event in ("change", "done"):
                self.assertEqual(changes, expected_changes)
            if event == "done":
                sysrepo.update_config_cache(current_config, changes)

        self.sess.subscribe_module_change_unsafe(
            "sysrepo-example",
            "/sysrepo-example:conf",
            module_change_cb,
            private_data=priv,
            update=True,
        )

        with self.conn.start_session("running") as ch_sess:
            # 1.
            sent_config = {"conf": {"system": {"hostname": "bar"}}}
            expected_changes = [
                sysrepo.ChangeCreated("/sysrepo-example:conf/system/hostname", "bar")
            ]
            ch_sess.replace_config(sent_config, "sysrepo-example", strict=True)
            self.assertEqual(current_config, sent_config)
            # 2.
            sent_config = {"conf": {"system": {"hostname": "INVALID"}}}
            expected_changes = []
            with self.assertRaises(sysrepo.SysrepoValidationFailedError):
                ch_sess.replace_config(sent_config, "sysrepo-example", strict=True)
            # 3.
            sent_config = {
                "conf": {
                    "system": {"hostname": "bar"},
                    "network": {"interface": [{"name": "eth0", "up": True}]},
                }
            }
            expected_changes = [
                sysrepo.ChangeCreated(
                    "/sysrepo-example:conf/network/interface[name='eth0']",
                    {"name": "eth0", "up": True},
                    after="",
                ),
                sysrepo.ChangeCreated(
                    "/sysrepo-example:conf/network/interface[name='eth0']/name", "eth0"
                ),
                sysrepo.ChangeCreated(
                    "/sysrepo-example:conf/network/interface[name='eth0']/up", True
                ),
            ]
            ch_sess.replace_config(sent_config, "sysrepo-example", strict=True)
            self.assertEqual(current_config, sent_config)
            # 4.
            sent_config = {
                "conf": {
                    "system": {"hostname": "bar"},
                    "network": {"interface": [{"name": "eth2", "up": False}]},
                }
            }
            expected_changes = [
                sysrepo.ChangeDeleted(
                    "/sysrepo-example:conf/network/interface[name='eth0']",
                    None,
                ),
                sysrepo.ChangeDeleted(
                    "/sysrepo-example:conf/network/interface[name='eth0']/name",
                    None,
                ),
                sysrepo.ChangeDeleted(
                    "/sysrepo-example:conf/network/interface[name='eth0']/up",
                    None,
                ),
                sysrepo.ChangeCreated(
                    "/sysrepo-example:conf/network/interface[name='eth2']",
                    {"name": "eth2", "up": False},
                    after="",
                ),
                sysrepo.ChangeCreated(
                    "/sysrepo-example:conf/network/interface[name='eth2']/name", "eth2"
                ),
                sysrepo.ChangeCreated(
                    "/sysrepo-example:conf/network/interface[name='eth2']/up", False
                ),
            ]
            ch_sess.replace_config(sent_config, "sysrepo-example", strict=True)
            self.assertEqual(current_config, sent_config)
            # 5.
            sent_config = {
                "conf": {
                    "system": {"hostname": "bar"},
                    "network": {"interface": [{"name": "eth2", "up": True}]},
                }
            }
            expected_changes = [
                sysrepo.ChangeModified(
                    "/sysrepo-example:conf/network/interface[name='eth2']/up",
                    True,
                    "false",
                    False,
                ),
            ]
            ch_sess.replace_config(sent_config, "sysrepo-example", strict=True)
            self.assertEqual(current_config, sent_config)
            # 6.
            sent_config = {
                "conf": {
                    "system": {"hostname": "bar"},
                    "network": {
                        "interface": [
                            {"name": "eth2", "up": True},
                            {"name": "eth0", "up": False},
                        ]
                    },
                }
            }
            expected_changes = [
                sysrepo.ChangeCreated(
                    "/sysrepo-example:conf/network/interface[name='eth0']",
                    {"name": "eth0", "up": False},
                    after="[name='eth2']",
                ),
                sysrepo.ChangeCreated(
                    "/sysrepo-example:conf/network/interface[name='eth0']/name", "eth0"
                ),
                sysrepo.ChangeCreated(
                    "/sysrepo-example:conf/network/interface[name='eth0']/up", False
                ),
            ]
            ch_sess.replace_config(sent_config, "sysrepo-example", strict=True)
            self.assertEqual(current_config, sent_config)
            # 7.
            sent_config = {
                "conf": {
                    "system": {"hostname": "bar"},
                    "network": {
                        "interface": [
                            {"name": "eth0", "up": False},
                            {"name": "eth2", "up": True},
                        ]
                    },
                }
            }
            expected_changes = [
                sysrepo.ChangeMoved(
                    "/sysrepo-example:conf/network/interface[name='eth0']", after=""
                ),
            ]
            ch_sess.replace_config(sent_config, "sysrepo-example", strict=True)
            self.assertEqual(current_config, sent_config)
